#  labcams - https://jpcouto@bitbucket.org/jpcouto/labcams.git
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from ximea import xiapi
from .cams import *

class XimeaCam(GenericCam):
    def __init__(self,
                 cam_id = None,
                 start_trigger = None,
                 stop_trigger = None,
                 save_trigger = None,                 
                 out_q = None,
                 binning = 2,
                 exposure = 20000,
                 triggerSource = np.uint16(2),
                 outputs = ['XI_GPO_EXPOSURE_ACTIVE'],
                 hardware_trigger = None,
                 recorderpar = None,
                 **kwargs):
        super(XimeaCam,self).__init__(cam_id = cam_id,
                                      out_q = out_q,
                                      start_trigger = None,
                                      stop_trigger = None,
                                      save_trigger = None,
                                      recorderpar = recorderpar)
        self.drivername = 'Ximea'
        self.hardware_trigger = hardware_trigger
        if self.hardware_trigger is None:
            self.hardware_trigger = Event()
        self.outputs = outputs
        self.binning = binning
        self.exposure = exposure
        self.frame_rate = 1000./float(self.exposure)
        self.fs.value = self.frame_rate

        frame = self.get_one()

        self.h = frame.shape[0]
        self.w = frame.shape[1]
        self.nchannels = 1
        self.dtype = np.uint16
        self._init_variables(self.dtype)
        self.triggerSource = triggerSource
        self.img[:] = np.reshape(frame,self.img.shape)[:]

        display("[Ximea {0}] - got info from camera.".format(self.cam_id))

    def _init_controls(self):
        self.ctrevents = dict(
            exposure=dict(
                function = 'set_exposure',
                widget = 'float',
                variable = 'exposure',
                units = 'ms',
                type = 'int',
                min = 1,
                max = 10000000000,
                step = 10))
    
    def get_one(self):
        self._cam_init()
        self.cam.start_acquisition()
        self.cam.get_image(self.cambuf)
        frame = self.cambuf.get_image_data_numpy()
        self.cam.stop_acquisition()
        self.cam.close_device()
        self.cam = None
        self.cambuf = None
        return frame
    
    def set_exposure(self,exposure = 20000):
        '''Set the exposure time is in us'''
        self.exposure = exposure
        self.frame_rate = 1000./float(self.exposure)
        self.fs.value = self.frame_rate

        if not self.cam is None:
            if self.cam_is_running:
                self.start_trigger.set()
                self.stop_trigger.set()

    def _cam_init(self,set_gpio=True):
        self.cam = xiapi.Camera()
        #start communication
        self.cam.open_device()
        self.cam.set_acq_timing_mode('XI_ACQ_TIMING_MODE_FREE_RUN')
        self.cam.set_exposure(self.exposure)
        self.cam.set_binning_vertical(self.binning)
        self.cam.set_binning_horizontal(self.binning)
        # Set the GPIO
        self.cam.set_gpi_selector('XI_GPI_PORT1')
        self.cam.set_trigger_source('XI_TRG_OFF')

        if set_gpio:
            self.cam.set_gpo_selector('XI_GPO_PORT1')
            # for now set the GPO to blink in software (port1 will do that for sync purposes, the test cam does not support XI_GPO_EXPOSURE_PULSE)
            self.cam.set_gpo_mode('XI_GPO_ON'); #XI_GPO_EXPOSURE_PULSE
            if self.hardware_trigger.is_set():
                self.cam.set_gpi_mode('XI_GPI_TRIGGER')
                self.cam.set_trigger_source('XI_TRG_EDGE_RISING')
        self.cam.set_led_selector('XI_LED_SEL1')
        self.cam.set_led_mode('XI_LED_OFF')
        self.cam.set_led_selector('XI_LED_SEL2')
        self.cam.set_led_mode('XI_LED_OFF')
        self.cam.set_led_selector('XI_LED_SEL3')
        self.cam.set_led_mode('XI_LED_OFF')
        self.cambuf = xiapi.Image()
        self.lastframeid = -1
        self.nframes.value = 0
        self.camera_ready.set()

    def _cam_startacquisition(self):
        self.cam.start_acquisition()

    def _cam_loop(self):
        try:
            self.cam.get_image(self.cambuf)
        except xiapi.Xi_error:
            return None,(None,None)
        frame = self.cambuf.get_image_data_numpy()
        frameID = self.cambuf.nframe
        timestamp = self.cambuf.tsUSec
        self.cam.set_gpo_mode('XI_GPO_OFF'); #XI_GPO_EXPOSURE_PULSE
        time.sleep(0.001)
        self.cam.set_gpo_mode('XI_GPO_ON'); #XI_GPO_EXPOSURE_PULSE
        return frame,(frameID,timestamp)
        
    def _cam_close(self):
        self.cam.stop_acquisition()
        self.cam.close_device()
        self.cam = None        



#  labcams - https://jpcouto@bitbucket.org/jpcouto/labcams.git
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtWidgets import (QWidget,
                             QApplication,
                             QGridLayout,
                             QFormLayout,
                             QVBoxLayout,
                             QHBoxLayout,
                             QTabWidget,
                             QCheckBox,
                             QTextEdit,
                             QLineEdit,
                             QComboBox,
                             QSlider,
                             QPushButton,
                             QLabel,
                             QAction,
                             QWidgetAction,
                             QMenuBar,
                             QDoubleSpinBox,
                             QGraphicsView,
                             QGraphicsScene,
                             QGraphicsItem,
                             QGraphicsLineItem,
                             QGroupBox,
                             QTableWidget,
                             QMainWindow,
                             QDockWidget,
                             QFileDialog,
                             QDialog,
                             QInputDialog)
from PyQt5.QtGui import QImage, QPixmap,QBrush,QPen,QColor,QFont
from PyQt5.QtCore import Qt,QSize,QRectF,QLineF,QPointF,QTimer

import pyqtgraph as pg
pg.setConfigOption('background', [200,200,200])
pg.setConfigOptions(imageAxisOrder='row-major')
pg.setConfigOption('crashWarning', True)

from .utils import *
cv2.setNumThreads(1)

colors = ['#d62728',
          '#1f77b4',
          '#ff7f0e',
          '#2ca02c',
          '#9467bd',
          '#8c564b',
          '#e377c2',
          '#7f7f7f',
          '#bcbd22'] 
penwidth = 1.


def sleep(dur = 1):
    tstart = time.time()
    while not (time.time()-tstart) > dur:
        QApplication.processEvents()

class QActionCheckBox(QWidgetAction):
    ''' Check box for the right mouse button dropdown menu'''
    def __init__(self,parent,label='',value=True):
        super(QActionCheckBox,self).__init__(parent)
        self.subw = QWidget()
        self.sublay = QFormLayout()
        self.checkbox = QCheckBox()
        self.sublab = QLabel(label)
        self.sublay.addRow(self.checkbox,self.sublab)
        self.subw.setLayout(self.sublay)
        self.setDefaultWidget(self.subw)
        self.checkbox.setChecked(value)
        self.value = self.checkbox.isChecked
    def link(self,func):
        self.checkbox.stateChanged.connect(func)

class QActionSlider(QWidgetAction):
    ''' Slider for the right mouse button dropdown menu'''
    def __init__(self,parent,label='',value=0,vmin = 0,vmax = 1000,step = None):
        super(QActionSlider,self).__init__(parent)
        self.subw = QWidget()
        self.sublay = QFormLayout()
        self.slider = QSlider()
        self.sublab = QLabel(label)
        self.sublay.addRow(self.sublab,self.slider)
        self.slider.setOrientation(Qt.Horizontal)
        self.subw.setLayout(self.sublay)
        self.setDefaultWidget(self.subw)
        self.slider.setMaximum(vmax)
        self.slider.setValue(value)
        if not step is None:
            self.slider.setSingleStep(step)
        self.slider.setMinimum(vmin)
        self.value = self.slider.value
    def link(self,func):
        self.slider.valueChanged.connect(func)

class QActionFloat(QWidgetAction):
    ''' Float edit for the right mouse button dropdown menu'''
    def __init__(self,parent,label='',value=0,
                 vmax = None,
                 vmin = None,
                 step = None):
        super(QActionFloat,self).__init__(parent)
        self.subw = QWidget()
        self.sublay = QFormLayout()
        self.spin = QDoubleSpinBox()
        self.sublab = QLabel(label)
        self.sublay.addRow(self.sublab,self.spin)
        self.subw.setLayout(self.sublay)
        self.setDefaultWidget(self.subw)
        if not vmax is None:
            self.spin.setMaximum(vmax)
        if not step is None:
            self.spin.setSingleStep(step)
        if not value is None:
            self.spin.setValue(value)
        if not vmin is None:
            self.spin.setMinimum(vmin)
        self.value = self.spin.value
    def link(self,func):
        self.spin.editingFinished.connect(func)

class RecordingControlWidget(QWidget):
    def __init__(self,parent):
        super(RecordingControlWidget,self).__init__()	
        self.parent = parent
        form = QFormLayout()

        info = '''Set the name of the experiment.
        Datapath is relative to the folder specified in the preference.
        Can be set via UDP (expname=my_experiment/name) or ZMQ (dict(action='expname',value='my_experiment/name'))
'''
        self.expname = ''
        w2 = QWidget()
        l2 = QFormLayout()
        w2.setLayout(l2)
        self.experimentNameEdit = QLineEdit(' ')
        self.experimentNameEdit.returnPressed.connect(self.checkUpdateFilename)
        label = QLabel('Name:')
        label.setToolTip(info)
        self.experimentNameEdit.setToolTip(info)
        l2.addRow(label,self.experimentNameEdit)
        w = QGroupBox('Triggers and acquisition')
        l = QFormLayout()
        w.setLayout(l)
        self.camTriggerToggle = QCheckBox()
        self.camTriggerToggle.setChecked(self.parent.hardware_trigger_event.is_set())
        self.camTriggerToggle.stateChanged.connect(self.toggleHardwareTriggered)
        label = QLabel("Hardware trigger: ")
        label.setToolTip(info)
        info = '''Toggle the hardware trigger mode on cameras that support it.
This will can be differently configured for different cameras.'''
        self.camTriggerToggle.setToolTip(info)
        #l.addRow(label,self.camTriggerToggle) # skipping harware trigger for now.

        self.saveButton = QPushButton('Acquire')
        info = '''The button turns off the camera, starts saving and starts the camera. 
This is the same as pressing the software trigger checkbox, then manual save then the software trigger again.
If the camera is saving this stops the camera.'''
        self.saveButton.setToolTip(info)
        self.snapshotButton = QPushButton('Snapshot')
        def snapshot():
            self.checkUpdateFilename()
            for cam,wid in zip(self.parent.cams,
                                      self.parent.camwidgets):
                if not cam.writer is None:
                    fname = cam.writer.get_filename_path()
                    dataname = cam.writer.dataname
                elif hasattr(cam,'recorder') and not cam.recorder is None:
                    fname = cam.recorder.get_filename_path()
                    dataname = cam.recorder.dataname
                else:
                    fname = None
                    dataname = 'snapshot'
                if not fname is None:
                    fname = pjoin(os.path.dirname(fname),'snapshots',
                                  datetime.today().strftime(
                                      '%Y%m%d_%H%M%S_{0}.tif'.format(dataname)))
                    display('[Snapshot] - {0}'.format(fname))
                    wid.saveImageFromCamera(filename=fname)
        self.snapshotButton.clicked.connect(snapshot)
        self.saveOnStartToggle = QCheckBox()
        self.softTriggerToggle = QCheckBox()
        def saveButton():
            update_shared_date()
            if not self.saveOnStartToggle.isChecked():
                self.softTriggerToggle.setChecked(False)
                sleep(0.1)
                self.saveOnStartToggle.setChecked(True)
                sleep(0.1)
                self.softTriggerToggle.setChecked(True)
                self.saveButton.setText('Stop')                
            else:                
                self.softTriggerToggle.setChecked(False)
                sleep(0.1)
                self.saveOnStartToggle.setChecked(False)
                self.saveButton.setText('Acquire')
        self.saveButton.clicked.connect(saveButton)
        #self.saveButton
        self.softTriggerToggle.setChecked(False)
        self.softTriggerToggle.stateChanged.connect(
            self.toggleSoftwareTriggered)
        label = QLabel("Software trigger: ")
        info = '''Toggle the software trigger to start or stop acquisition using the camera control software.'''
        label.setToolTip(info)
        self.softTriggerToggle.setToolTip(info)
        l.addRow(label,self.softTriggerToggle)

        self.saveOnStartToggle.setChecked(self.parent.save_on_start)
        self.saveOnStartToggle.stateChanged.connect(self.toggleSaveOnStart)

        label = QLabel("Manual save: ")
        info = '''Stream data to disk. The checkbox starts and stops acquisition.'''
        label.setToolTip(info)
        self.saveOnStartToggle.setToolTip(info)
        self.saveButton.setFixedWidth(100)
        l.addRow(label,self.saveOnStartToggle)

        self.udpmessages = QLabel('')
        b1=QFont()
        b1.setPixelSize(16)
        b1.setFamily('Regular')
        b1.setBold(True)
        self.udpmessages.setFont(b1)
        self.udpmessages.setStyleSheet("color: rgb(255,165,0)")
        form.addRow(self.udpmessages)
        self.setLayout(form)
        self.layout = form
        l.addRow(self.saveButton,self.snapshotButton)
        l2.addRow(self.udpmessages)
        form.addRow(w,w2)

    def checkUpdateFilename(self,filename=None):
        if filename is None:
            filename = str(self.experimentNameEdit.text())
        if not self.expname == filename:
            self.parent.set_experiment_name(filename)
            self.expname = filename
        
    def toggleSoftwareTriggered(self,value):
        display('[labcams] Software trigger [{0}]'.format(value))
        if value:
            for cam in self.parent.cams:
                cam.start_acquisition()
                if hasattr(cam.cam,'analog_channels'):
                    display('Sleeping for 1 second for the DAQ to record.')
                    sleep(1)
            tstart[0] = time.time()
        else:
            for cam in self.parent.cams[::-1]:
                if hasattr(cam.cam,'analog_channels'):
                    display('Sleeping for 1 second for the DAQ to record.')
                    sleep(1)
                cam.stop_acquisition()
            camready = 0
            while camready != len(self.parent.cams):
                camready = np.sum([cam.camera_ready.is_set() for cam in self.parent.cams])
                QApplication.processEvents()
            display('[labcams] - All cameras ready to be triggered.')


    def toggleHardwareTriggered(self,value):
        update_shared_date()
        display('Hardware trigger mode pressed [{0}]'.format(value))
        if self.parent.save_on_start:
            self.checkUpdateFilename()
        if value:
            self.parent.hardware_trigger_event.set()
            self.recController.softTriggerToggle.setChecked(True)
            self.recController.saveOnStartToggle.setChecked(True)
            tstart[0] = time.time()
        else:
            #self.toggleSaveOnStart(False)
            # save button does not get unticked (this is a bug)
            self.parent.hardware_trigger_event.clear()
            self.recController.softTriggerToggle.setChecked(False)
            self.recController.saveOnStartToggle.setChecked(False)
        
    def toggleSaveOnStart(self,state):
        update_shared_date()
        self.parent.save_on_start = state
        self.checkUpdateFilename()
        for c,(cam,flg) in enumerate(zip(self.parent.cams,
                                                self.parent.saveflags)):
            if flg:
                cam.set_saving(state)
                if state:
                    self.experimentNameEdit.setDisabled(True)
                else:
                    self.experimentNameEdit.setDisabled(False)
        
class CamWidget(QWidget):
    def __init__(self,
                 frame,
                 iCam = 0,
                 parent = None,
                 parameters = None,
                 invertX = False):
        super(CamWidget,self).__init__()
        self.parent = parent
        self.iCam = iCam
        self.cam = self.parent.cams[self.iCam]
        self.nchan = frame.shape[-1]
        if hasattr(self.cam,'excitation_trigger'):
            self.nchan = self.cam.excitation_trigger.nchannels.value
            self.frame_buffer = None
        self.displaychannel = -1  # default show all channels
        self.roiwidget = None
        self.layout = QGridLayout()
        self.setLayout(self.layout)
        win = pg.GraphicsLayoutWidget()
        p1 = win.addPlot(title="")
        self.view = pg.ImageItem(background=[1,1,1])
        p1.getViewBox().invertY(True)
        if invertX:
            p1.getViewBox().invertX(True)
        p1.getViewBox().setAspectLocked(True)
        p1.hideAxis('left')
        p1.hideAxis('bottom')
        p1.addItem(self.view)
        self.text = pg.TextItem('',color = [220,80,80],anchor = [0,0])
        p1.addItem(self.text)
        b=QFont()
        b.setPixelSize(14)
        b.setFamily('Regular')
        b.setBold(False)
        self.text.setFont(b)
        # remotemsg
        #self.text_remote = pg.TextItem('',color = [220,100,200],anchor = [0,0.1])
        #p1.addItem(self.text_remote)
        #b1=QFont()
        #b1.setPixelSize(14)
        #b1.setFamily('Regular')
        #b1.setBold(False)
        #self.text_remote.setFont(b1)
        
        self.layout.addWidget(win,0,0)
        self.p1 = p1
        self.autoRange = True
        self.lastnFrame = 0
        if not 'SubtractBackground' in parameters.keys():
            parameters['SubtractBackground'] = False
        if not 'Equalize' in parameters.keys():
            parameters['Equalize'] = False
        if not 'TrackEye' in parameters.keys():
            parameters['TrackEye'] = False
        self.parameters = parameters
        self.parameters['reference_channel'] = None
        self.lastFrame = 0
        if not 'NBackgroundFrames' in parameters.keys() or not parameters['SubtractBackground']:
            self.nAcum = 0
        else:
            self.nAcum = float(parameters['NBackgroundFrames'])
        self.eyeTracker = None
        self.string = '{0}'
        if not self.parameters['save_data']:
            self.string = 'no save -{0}'
        #self.image(np.array(frame),-1)
        size = 600
        #ratio = h/float(w)
        self.hist = None
###        self.setFixedSize(size,int(size*ratio))
        #self.resize(size,int(size*ratio))
        self.addActions()

        #self.show()
    def update(self):
        # handle the excitation module
        if hasattr(self.cam,'excitation_trigger'):
            if self.frame_buffer is None:
                self.frame_buffer = np.zeros([
                    self.cam.cam.h.value,
                    self.cam.cam.w.value,
                    3], dtype = self.cam.cam.dtype)
            cframe = self.cam.nframes.value
            tmp = self.cam.get_img(cframe)
            nchan = self.cam.excitation_trigger.nchannels.value
            self.frame_buffer[:,:,
                              np.mod(cframe,
                                     nchan)] = tmp.squeeze()
            return self.image(self.frame_buffer,cframe)
        else:
            frame = self.cam.get_img()
        if not frame is None:
            self.image(frame,self.cam.nframes.value)
            
    def addActions(self):
        self.setContextMenuPolicy(Qt.ActionsContextMenu)
        saveImg = QAction("Take camera shot",self)
        saveImg.triggered.connect(self.saveImageFromCamera)
        self.addAction(saveImg)
        sep = QAction(self)
        sep.setSeparator(True)
        self.addAction(sep)
        self.functs = []
        # add camera controls
        if hasattr(self.cam.cam,'ctrevents'):
            self.ctract = dict()
            def vchanged(the):
                val = the['action'].value()
                self.cam.cam.eventsQ.put(the['name']+'='+str(val))

            for k in  self.cam.cam.ctrevents.keys():
                self.ctract[k] = dict(**self.cam.cam.ctrevents[k])
                ev = self.ctract[k]
                val = eval('self.cam.cam.' + ev['variable'])
                ev['name'] = k
                ev['action'] = None
                if ev['widget'] == 'slider':
                    ev['action'] = QActionSlider(self,
                                                 k+' [{0:03d}]:'.format(int(val)),
                                                 value = val,
                                                 vmin = ev['min'],
                                                 vmax = ev['max'],
                                                 step = ev['step'])
                elif ev['widget'] == 'float':
                    ev['action'] = QActionFloat(self,k,
                                                value = val,
                                                vmin = ev['min'],
                                                vmax = ev['max'],
                                                step = ev['step'])
                    
                if not ev['action'] is None:
                        #e.sublab.setText(k + ' [{0:03d}]:'.format(int(val)))
                    self.functs.append(partial(vchanged,self.ctract[k]))
                    ev['action'].link(self.functs[-1]) 
                    self.addAction(ev['action'])
            
        sep = QAction(self)
        sep.setSeparator(True)
        self.addAction(sep)
        if self.nchan > 1:
            displaychan = QActionSlider(self,'display channel',
                                       value = self.displaychannel,
                                       vmin = -1,
                                       vmax = self.nchan-1)
            def change_chan(value):
                self.displaychannel = int(value)
                if not self.roiwidget is None:
                    self.roiwidget.reset()
            displaychan.link(change_chan) 
            self.addAction(displaychan)
        # Slider
        toggleSubtract = QActionSlider(self,'Nsubtract [{0:03d}]:'.format(int(self.nAcum)),
                                       value = 0,
                                       vmin = 0,
                                       vmax = 200)
        toggleSubtract.link(lambda x: toggleSubtract.sublab.setText(
            'Nsubtract [{0:03d}]:'.format(int(x))))
        def vchanged(val):
            self.nAcum = float(np.floor(val))
            if not self.roiwidget is None:
                self.roiwidget.reset()

        toggleSubtract.link(vchanged) 
        # ROIs
        self.addAction(toggleSubtract)
        addroi = QAction("Add ROI",self)
        addroi.triggered.connect(self.addROI)
        self.addAction(addroi)
        # Equalize histogram
        toggleEqualize = QActionCheckBox(self,'Equalize histogram',self.parameters['Equalize'])
        def toggleEq():
            self.parameters['Equalize'] = not self.parameters['Equalize']
            toggleEqualize.checkbox.setChecked(self.parameters['Equalize'])
        toggleEqualize.link(toggleEq)
        self.addAction(toggleEqualize)
        # Eye tracker
        self.etrackercheck = QActionCheckBox(self,'Eye tracker',
                                             self.parameters['TrackEye'])
        self.etrackercheck.link(self.toggleEyeTracker)
        self.addAction(self.etrackercheck)
        # autorange
        
        self.autorange_toggle = QActionCheckBox(self,'Auto range',self.autoRange)
            
        self.autorange_toggle.link(self.toggleAutoRange)
        self.addAction(self.autorange_toggle)
        
        # histogram
        tEt = QAction('Histogram',self)
        tEt.triggered.connect(self.histogramWin)
        self.addAction(tEt)

        tEt = QAction('Saturation mode',self)
        tEt.triggered.connect(self.saturationMode)
        self.addAction(tEt)
        
        # Save
        ts = QActionCheckBox(self,'Save camera',  self.parent.saveflags[self.iCam])
        def toggleSaveCam():
            self.parameters['save_data'] = not self.parameters['save_data']
            self.parent.saveflags[self.iCam] = self.parameters['save_data']
            ts.checkbox.setChecked(self.parameters['save_data'])
            if not self.parameters['save_data']:
                self.string = 'no save - {0}'
            else:
                self.string = '{0}'            
        ts.link(toggleSaveCam)
        self.addAction(ts)
        self.reference_toggle = QActionCheckBox(self,'alignment reference',  False)
        self.reference_toggle.link(self.toggle_reference)
        self.addAction(self.reference_toggle)
        
    def toggle_reference(self,filename):
        if self.parameters['reference_channel'] is None:
            if not type(filename) is str:
                fdlg = QFileDialog().getOpenFileName(
                    self,
                    'Load reference image')
                filename = str(fdlg[0])

                print('Selected {0}'.format(filename))
            else:

                self.reference_toggle.checkbox.disconnect()
                self.reference_toggle.checkbox.setChecked(True)
                self.reference_toggle.link(self.toggle_reference)
                self.reference_toggle.value = True
        else:
            self.reference_toggle.checkbox.disconnect()
            self.reference_toggle.checkbox.setChecked(False)
            self.reference_toggle.link(self.toggle_reference)
            self.reference_toggle.value = False
            self.parameters['reference_channel'] = None
            return        
        if type(filename) is str and os.path.exists(filename):
            try:
                from skimage.io import imread
                from skimage.transform import resize
            except:
                from tifffile import imread
            display(filename)
            reference = imread(filename).squeeze()
            if len(reference.shape) > 2:
                reference = reference.mean(axis = -1)
            if 'resize' in dir(): # reshape if possible/needed
                reference = resize(reference,
                                   output_shape = self.view.image.shape)
            self.parameters['reference_channel'] = reference
            self.reference_toggle.checkbox.setChecked(True)

    def toggleAutoRange(self,value):
        self.autoRange = value #not self.autoRange
        if not self.autorange_toggle.checkbox.isChecked == value:
            self.autorange_toggle.checkbox.setChecked(self.autoRange)
        
    def histogramWin(self):
        if self.hist is None:
            histTab = QDockWidget("histogram cam {0}".format(self.iCam), self)
            widget = QWidget()
            layout = QGridLayout()
            widget.setLayout(layout)
            win = pg.GraphicsLayoutWidget()
            p1 = win.addPlot()
            p1.getViewBox().invertY(True)
            p1.hideAxis('left')
            p1.hideAxis('bottom')
            
            self.hist = pg.HistogramLUTItem()
            self.hist.axis.setPen('k')
            p1.addItem(self.hist)
            self.hist.setImageItem(self.view)
            layout.addWidget(win,0,0)
            histTab.setWidget(widget)
            histTab.setAllowedAreas(Qt.LeftDockWidgetArea |
                                    Qt.RightDockWidgetArea |
                                    Qt.BottomDockWidgetArea |
                                    Qt.TopDockWidgetArea)
            histTab.setFeatures(QDockWidget.DockWidgetMovable |
                                QDockWidget.DockWidgetFloatable |
                                QDockWidget.DockWidgetClosable)
            self.parent.addDockWidget(Qt.BottomDockWidgetArea
                                      ,histTab)
            histTab.setFloating(True)
        
            histTab.resize(200,200)
        try:
            histstate = self.hist.saveState()
        except Exception as err:
            display('[ERROR] - could not save histogram state. "pip install pyqtgraph --upgrade" might solve it.')
            print(err)
        def closefun(ev):
            try:
                self.hist.restoreState(histstate)
            except Exception as err:
                print(err)
                pass
            self.hist = None
            ev.accept()
        histTab.closeEvent = closefun
        
    def saturationMode(self):
        if self.hist is None:
            self.histogramWin()
        if not self.hist.gradient is None:
            self.hist.gradient.addTick(0.751, color=pg.mkColor('#ff2d00'))
            self.hist.gradient.addTick(1, color=pg.mkColor('#ff2d00'))
            self.hist.gradient.addTick(0.75, color=pg.mkColor('#ffffff'))
            self.toggleAutoRange(False)
            dt = self.view.image.dtype
            self.hist.setLevels(np.iinfo(dt).min,np.iinfo(dt).max)

        
    def addROI(self,roi = None):
        if self.roiwidget is None:
            self.roiwidget = ROIPlotWidget(roi_target = self.p1,
                                           view = self.view,
                                           parent = self)
            roiTab = QDockWidget("roi cam {0}".format(self.iCam), self)
            roiTab.setWidget(self.roiwidget)
            roiTab.setAllowedAreas(Qt.LeftDockWidgetArea |
                                   Qt.RightDockWidgetArea |
                                   Qt.BottomDockWidgetArea |
                                   Qt.TopDockWidgetArea)
            roiTab.setFeatures(QDockWidget.DockWidgetMovable |
                               QDockWidget.DockWidgetFloatable |
                               QDockWidget.DockWidgetClosable)
            self.parent.addDockWidget(Qt.BottomDockWidgetArea
                                      ,roiTab)
            roiTab.setFloating(True)
            roiTab.resize(600,150)
            
            def closetab(ev):
                # This probably does not clean up memory...
                if not self.roiwidget is None:
                    [self.p1.removeItem(r)
                     for r in self.roiwidget.items()]
                    del self.roiwidget
                    self.roiwidget = None
                ev.accept()
            roiTab.closeEvent = closetab
        self.roiwidget.add_roi(roi)

    def toggleEyeTracker(self):
        if self.parameters['TrackEye']:
            self.eyeTracker = None
            if hasattr(self,'trackerpar'):
                self.trackerpar.close()
                self.trackerTab.close()
                [self.p1.removeItem(c) for c in self.tracker_roi.items()]
        self.parameters['TrackEye'] = not self.parameters['TrackEye']
        self.etrackercheck.checkbox.setChecked(self.parameters['TrackEye'])

    def saveImageFromCamera(self,filename=None):
        update_shared_date()
        frame = self.parent.cams[self.iCam].get_img_with_virtual_channels()
        
        if filename is None:
            self.parent.timer.stop()
            filename = QFileDialog.getSaveFileName(self,
                                               'Select filename to save.')
            self.parent.timer.start()
        if type(filename) is tuple:
            filename = filename[0]
        if filename:
            folder = os.path.dirname(filename)
            if not os.path.isdir(folder):
                os.makedirs(folder)
            from tifffile import imsave
            imsave(str(filename),
                   frame.transpose([2,0,1]).squeeze(),
                   metadata = {
                       'Camera':str(self.iCam)})
            display('Saved camera frame for cam: {0}'.format(self.iCam))
        else:
            display('Snapshot aborted.')
        
    def _open_mptracker(self,image):
        try:
            from mptracker import MPTracker
            from mptracker.widgets import MptrackerParameters,EyeROIWidget
        except:
            display('Could not load tracker.')
            display('\nInstall mptracker: https://bitbucket.org/jpcouto/mptracker')
            return
        self.eyeTracker = MPTracker(drawProcessedFrame=True)
        self.trackerTab = QDockWidget("mptracker cam {0}".format(self.iCam), self)
        self.eyeTracker.parameters['crTrack'] = True
        self.eyeTracker.parameters['sequentialCRMode'] = False
        self.eyeTracker.parameters['sequentialPupilMode'] = False
        self.tracker_roi = EyeROIWidget()
        [self.p1.addItem(c) for c in  self.tracker_roi.items()]
        self.trackerpar = MptrackerParameters(self.eyeTracker,image,eyewidget=self.tracker_roi)
        if self.parent.saveflags[self.iCam]:
            self.trackerToggle = QCheckBox()
            if not self.parent.cams[self.iCam].writer is None:
                pass
                #self.trackerToggle.setChecked(self.parent.writers[self.iCam].trackerFlag.is_set())
            #self.trackerToggle.stateChanged.connect(self.trackerSaveToggle)
            self.trackerpar.pGridSave.addRow(
                QLabel("Save cameras: "),self.trackerToggle)
        self.restracker = [0]
        self.addROI(self.restracker)
        self.trackerTab.setWidget(self.trackerpar)
        self.trackerTab.setFloating(True)
        self.trackerpar.resize(400,250)
        self.trackerTab.setAllowedAreas(Qt.LeftDockWidgetArea |
                                        Qt.LeftDockWidgetArea |
                                        Qt.BottomDockWidgetArea |
                                        Qt.TopDockWidgetArea )
        self.trackerTab.setFeatures(QDockWidget.DockWidgetMovable |
                                  QDockWidget.DockWidgetFloatable |
                                  QDockWidget.DockWidgetClosable)
        self.parent.addDockWidget(Qt.LeftDockWidgetArea
                                  ,self.trackerTab)
    def trackerSaveToggle(self,value):
        writer = self.parent.cams[self.iCam].writer
        if not writer is None:
            if self.parent.saveflags[self.iCam]:
                if value:
                    writer.trackerFlag.set()
                    writer.parQ.put((None,self.eyeTracker.parameters))
                else:
                    writer.trackerFlag.clear()

    def image(self,image,nframe):
        if self.lastnFrame != nframe:
            tmp = image.copy()                    
            if self.parameters['Equalize']: # histogram equalization
                try: 
                    tmp = cv2.equalizeHist(tmp).reshape(image.shape)
                except:
                    pass
            if self.nAcum > 0:  # subtraction 
                tmp = tmp.astype(np.float32)
                frame = (tmp - self.lastFrame)
                self.lastFrame = ((1.-1./self.nAcum)*self.lastFrame +
                                  (1./self.nAcum)*tmp)
            else:
                self.lastFrame = 0
                frame = tmp
            if not self.roiwidget is None:
                self.roiwidget.update(frame,iFrame=nframe)
            if bool(self.parameters['TrackEye']):
                if self.eyeTracker is None:
                    self._open_mptracker(image[:,:,0])
                self.restracker[0] = np.nanmax(self.eyeTracker.apply(image[:,:,0])[2])
                self.displaychannel = -1
                if not self.eyeTracker.concatenateBinaryImage:
                    (x1,y1,w,h) = self.eyeTracker.parameters['imagecropidx']
                    if frame.shape[2] != 3:
                        frame = cv2.cvtColor(frame[:,:,0],cv2.COLOR_GRAY2RGB)
                    frame[y1:y1+h,x1:x1+w] = self.eyeTracker.img
                else:
                    frame = self.eyeTracker.img

            self.text.setText(self.string.format(nframe))
            if not self.displaychannel == -1:
                frame = frame[:,:,self.displaychannel]

            if self.parameters['reference_channel'] is None:
                if self.displaychannel == -1 and frame.shape[2]==2:
                    f = frame.copy()
                    frame = np.zeros((f.shape[0],f.shape[1],3),dtype = f.dtype)
                    frame[:,:,1] = f[:,:,0]
                    frame[:,:,2] = f[:,:,1]
                self.view.setImage(frame.squeeze(),
                                   autoLevels=self.autoRange,autoDownsample=True)
            else:
                frame = frame.squeeze()
                ref = self.parameters['reference_channel']
                im = np.stack([ref/np.max(ref),frame/np.max(frame),np.zeros_like(frame)]).transpose([1,2,0])
                self.view.setImage(im,autoDownsample=True)
            self.lastnFrame = nframe


class ROIPlotWidget(QWidget):
    def __init__(self, roi_target= None,view=None,npoints = 1200,parent = None):
        super(ROIPlotWidget,self).__init__()	
        layout = QGridLayout()
        self.parent=parent
        self.setLayout(layout)
        self.view = view
        self.roi_target = roi_target
        win = pg.GraphicsLayoutWidget()
        self.p1 = win.addPlot()
        self.p1.getAxis('bottom').setPen('k') 
        self.p1.getAxis('left').setPen('k') 
        layout.addWidget(win,0,0)
        self.N = npoints
        self.rois = []
        self.plots = []
        self.buffers = []

    def add_roi(self,roi = None):
        pencolor = colors[
            np.mod(len(self.plots),len(colors))]
        if not type(roi) is list:
            roi = pg.RectROI(pos=[100,100],
                             size=100,
                             pen=pencolor)
            self.roi_target.addItem(roi)
        self.rois.append(roi)
        self.plots.append(pg.PlotCurveItem(pen=pg.mkPen(
        color=pencolor, width=penwidth)))
        self.p1.addItem(self.plots[-1])
        buf = np.zeros([2,self.N],dtype=np.float32)
        buf[0,:] = np.nan
        buf[1,:] = 0
        self.buffers.append(buf)

    def items(self):
        return self.rois

    def closeEvent(self,ev):
        for roi in self.rois:
            self.roi_target.removeItem(roi)
        ev.accept()

    def reset(self):
        for ib in range(len(self.buffers)):
            self.buffers[ib][0,:] = np.nan
        
    def update(self,img,iFrame):
        
        ichan = -1
        if not self.parent is None:
            ichan = self.parent.displaychannel
        ctime = time.time() - tstart[0]
        if len(self.buffers):
            if (ctime  - np.nanmax(self.buffers[0][0,:])) < 0:
                print('resetting')
                self.reset()
        for i,(roi,plot) in enumerate(zip(self.rois,self.plots)):
            if type(roi) is list:
                r = np.array(roi)[-1]
            else:
                r = np.mean(roi.getArrayRegion(img[:,:,ichan], self.view)).copy()
            buf = np.roll(self.buffers[i], -1, axis = 1)
            buf[0,-1] = ctime
            buf[1,-1] = r
            self.buffers[i] = buf
            ii = (~np.isnan(buf[0,:])) & (~np.isnan(buf[1,:]))
            plot.setData(x = buf[0,ii],
                         y = buf[1,ii])

class CamStimTriggerWidget(QWidget):
    def __init__(self,port = None,ino=None, outQ = None, cam = None):
        super(CamStimTriggerWidget,self).__init__()
        if (ino is None) and (not port is None):
            from .cam_stim_trigger import CamStimInterface
            ino = CamStimInterface(port = port,outQ = outQ)
        self.ino = ino
        self.cam = cam
        form = QFormLayout()
        if not ino is None:
            def disarm():
                ino.disarm()
            disarmButton = QPushButton('Disarm')
            disarmButton.clicked.connect(disarm)
            form.addRow(disarmButton)

            def arm():
                ino.arm()
            armButton = QPushButton('Arm')
            armButton.clicked.connect(arm)
            form.addRow(armButton)

            if len(self.ino.modes):
                wcombo = QComboBox()
                wcombo.addItems(self.ino.modes)
                wcombo.setCurrentIndex(len(self.ino.modes)-1)
                wcombo.currentIndexChanged.connect(self.setMode)
                self.ino.set_mode(len(self.ino.modes))
                form.addRow(wcombo)

            wsync = QLabel()
            wsyncstr = 'sync {0} - frame {1}'
            def update_count():
                if ino.sync.value:
                    wsyncstr = '<b>sync {0} - frame {1}<\b>'
                else:
                    wsyncstr = 'sync {0} - frame {1}'               
                wsync.setText(wsyncstr.format(ino.sync_count.value,
                                              ino.sync_frame_count.value))
            
            form.addRow(wsync)
            self.t = QTimer()
            self.t.timeout.connect(update_count)
            self.t.start(100)
        self.setLayout(form)
            
    def setMode(self,i):
        self.ino.set_mode(i+1)
        self.ino.check_nchannels()
        sleep(0.1)
        if not self.cam is None:
            self.cam.nchan = self.ino.nchannels.value
        
    def close(self):
        self.ino.close()
        self.ino.join()


class SettingsDialog(QDialog):
    def __init__(self, settings = None):
        super(SettingsDialog,self).__init__()
        self.setWindowTitle('labcams')
        from .utils import _SERVER_SETTINGS,_RECORDER_SETTINGS,_OTHER_SETTINGS
        if settings is None:
            settings = {}
            for s in _SERVER_SETTINGS.keys():
                if not s.endswith('_help'):
                    if type(_SERVER_SETTINGS[s]) is list:
                        settings[s] = _SERVER_SETTINGS[s][0]
                    else:
                        settings[s] = _SERVER_SETTINGS[s]
                for s in _OTHER_SETTINGS.keys():
                    if not s.endswith('_help'):
                        settings[s] = _OTHER_SETTINGS[s]

            settings['cams'] = []
        self.currentcam = None    
        self.settings = settings
        layout = QFormLayout()
        self.setLayout(layout)

        from PyQt5.QtWidgets import QListWidget,QTabWidget
        
        self.cams_listw = QListWidget()
        for i,c in enumerate(settings['cams']):
            self.cams_listw.addItem('{0} - {1}'.format(c['name'],c['driver']))
            self.currentcam = i
        btadd = QPushButton('Add')
        layout.addRow('Cameras',btadd)
        btremove = QPushButton('Remove')
        nw = QWidget()
        nl = QHBoxLayout()
        nl.addWidget(btadd)
        nl.addWidget(btremove)
        nw.setLayout(nl)
        layout.addRow('Cameras',nw)
        def addcamera():
            self.settings['cams'].append(DEFAULTS['cams'][-1])
            c = self.settings['cams'][-1]
            self.cams_listw.addItem('{0} - {1}'.format(c['name'],
                                                       c['driver']))
        self.camwidget = CamSettingsDialog(settings = self.settings)

        def camselect():
            idx = self.cams_listw.currentRow()
            self.currentcam = idx
            self.camwidget.set_camera(idx)
            
        self.cams_listw.itemClicked.connect(camselect)
        btadd.clicked.connect(addcamera)
        b1 = QGroupBox()
        b1.setTitle('Remote (network) access settings')
        lay = QFormLayout(b1)
        for k in _SERVER_SETTINGS.keys():
            if not k.endswith('_help'):
                if not k in self.settings.keys():
                    self.settings[k] = _SERVER_SETTINGS[k]
                    if type(_SERVER_SETTINGS[k]) is list:
                        self.settings[k] = self.settings[k][0]
                if type(_SERVER_SETTINGS[k]) is list:
                    # then it is an option menu
                    par = QComboBox()
                    for i in _SERVER_SETTINGS[k]:
                        par.addItem(i)
                    print(self.settings[k])
                    index = par.findText(self.settings[k])
                    if index >= 0:
                        par.setCurrentIndex(index)
                else:
                    par = QLineEdit()
                    par.setText(str(self.settings[k]))
                lay.addRow(QLabel(k),par)
        layout.addRow(self.cams_listw,b1)
        layout.addRow(self.camwidget)
        b2 = QGroupBox()
        b2.setTitle('General settings')
        lay = QFormLayout(b2)
        for k in _OTHER_SETTINGS.keys():
            if not k.endswith('_help'):
                par = QLineEdit()
                if not k in self.settings.keys():
                    self.settings[k] = _OTHER_SETTINGS[k]
                par.setText(str(self.settings[k]))
                lay.addRow(QLabel(k),par)
        layout.addRow(b2)
        
        self.show()

class CamSettingsDialog(QWidget):
    def __init__(self, settings = None):
        super(CamSettingsDialog,self).__init__()
        if settings is None:
            settings = dict()
        self.settings = settings
        self.current = 0
        self.cam = dict()
        
        layout = QFormLayout()
        self.setLayout(layout)
        from .utils import _CAMERA_SETTINGS,_RECORDER_SETTINGS,_CAMERAS
        self.b1 = QGroupBox()
        self.b1.setTitle('Camera settings')
        self.b1_lay = QFormLayout(self.b1)
        self.drivername = QComboBox()
        for k in _CAMERAS.keys():
            self.drivername.addItem(_CAMERAS[k])
        self.b1_w = []
        w1 = QWidget()
        l = QFormLayout()
        w1.setLayout(l)
        l.addRow('Camera driver',self.drivername)
        self.drivername.currentIndexChanged.connect(self.set_driver)
        w2 = QWidget()
        l = QFormLayout()
        w2.setLayout(l)
        self.camname = QLineEdit()
        self.camname.setText('cam1')
        l.addRow('Name',self.camname)
        self.b1_lay.addRow(w1,w2)
        self.b2 = QGroupBox()
        self.b2.setTitle('Recorder settings')
        self.b2_lay = QFormLayout(self.b2)
        par = QComboBox()
        for i in _RECORDER_SETTINGS['recorder']:
            par.addItem(i)
        self.b2_w = []
        self.b2_lay.addRow('Recorder format',par)
        self.use_queue = QCheckBox()
        self.use_queue.setChecked(True)
        self.b2_lay.addRow('Use frame queue',self.use_queue)
        layout.addRow(self.b1)
        layout.addRow(self.b2)

    def set_camera(self,idx):
        self.current = idx
        self.camsettings = self.settings['cams'][self.current]
        self.set_driver(self.camsettings['driver'])
        
    def set_driver(self,value=None):
        from .utils import _CAMERA_SETTINGS,_CAMERAS
        drivers = [k for k in _CAMERAS.keys()]
        if not value is None:
            if type(value) is int:
                camdriver = drivers[value]
            else:
                camdriver = value.lower()
            if not len(self.camsettings):
                self.camsettings = dict(_CAMERA_SETTINGS[camdriver],driver = camdriver)
            # check that what is in the settings is not overwritten.
            self.drivername.setCurrentIndex(drivers.index(camdriver)) 
        self.camsettings['driver'] = camdriver
        self.set_camera_widgets()
        
    def set_camera_widgets(self):
        if len(self.b1_w):
            for i in self.b1_w:
                self.b1_lay.removeRow(i[0])
        self.b1_w = []
        from .utils import _CAMERA_SETTINGS,_CAMERAS
        sett = self.camsettings
        print(sett)
        for k in sett.keys():
            if not k in _CAMERA_SETTINGS[self.camsettings['driver']]:
                continue
            if type(sett[k]) is list:
                # then it is an option menu
                par = QComboBox()
                for i in sett[k]:
                    par.addItem(i)
                index = par.findText(self.settings[k])
                if index >= 0:
                    par.setCurrentIndex(index)
            else:
                par = QLineEdit()
            par.setText(str(self.camsettings[k]))
            l = QLabel(k)
            self.b1_lay.addRow(l,par)
            self.b1_w.append([l,par])

class DAQPlotWidget(QWidget):
    def __init__(self,
                 daq,
                 parent = None,
                 parameters = None):
        super(DAQPlotWidget,self).__init__()
        self.parent = parent
        self.daq = daq
        self.parameters = parameters
        layout = QGridLayout()
        self.setLayout(layout)

        win = pg.GraphicsLayoutWidget()
        self.p1 = win.addPlot()
        self.p1.getAxis('bottom').setPen('k') 
        self.p1.getAxis('left').setPen('k') 
        layout.addWidget(win,0,0)
        self.plots = []
        self.N = self.daq.data_buffer.shape[1]
        for i in range(self.daq.data_buffer.shape[0]):
            
            pencolor = colors[np.mod(i,len(colors))]
            self.plots.append(pg.PlotCurveItem(pen=pg.mkPen(
                color=pencolor, width=penwidth)))
            self.p1.addItem(self.plots[-1])

        chan = [self.daq.analog_channels[k] for k in self.daq.analog_channels.keys()]
        chan += [self.daq.digital_channels[k] for k in self.daq.digital_channels.keys()]
        self.p1.getAxis('left').setTicks([[
            ((1*i),p) for i,p in enumerate(chan)]])

    def update(self):
        for i, plot in enumerate(self.plots):
            Y = self.daq.data_buffer[i]
            if i<self.daq.ai_num_channels:
                # divide by int16:
                Y = Y.astype('float32')/32767
            plot.setData(x = np.arange(0,self.N),
                         y = Y*0.7 + i)



#  labcams - https://jpcouto@bitbucket.org/jpcouto/labcams.git
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .cams import *
# QImaging cameras
from .qimaging_dll import *


# I don't have access to a Rollera Emc2 Camera so can't test it.
# Please let me know if this gives issues.
class QImagingCam(GenericCam):
    def __init__(self,
                 cam_id = None,
                 start_trigger = None,
                 stop_trigger = None,
                 save_trigger = None,
                 out_q = None,
                 exposure = 100000,
                 gain = 3500,frame_timeout = 100,
                 n_frame_buffers = 1,
                 binning = 2,
                 trigger_type = 0,
                 hardware_trigger = Event(),
                 recorderpar = None):
        '''
        Qimaging camera (tested with the Emc2 only!)
            trigger_type (0=freerun,1=hardware,5=software)
        '''
        super(QImagingCam,self).__init__(cam_id = cam_id,
                                         out_q = out_q,
                                         start_trigger = start_trigger,
                                         stop_trigger = stop_trigger,
                                         save_trigger = save_trigger,    
                                         recorderpar = recorderpar)
        self.hardware_trigger = hardware_trigger
        if self.hardware_trigger is None:
            self.hardware_trigger = Event()
        self.trigger_type = 0
        self.estimated_readout_lag = 1257 # microseconds
        self.binning = binning
        self.exposure = exposure
        self.gain = gain
        self.frame_rate = 1./(self.exposure/1000.)
        self.fs.value = self.frame_rate

        self.frame_timeout = frame_timeout
        self.nbuffers = n_frame_buffers
        self.trigger_type = trigger_type
        ReleaseDriver()
        LoadDriver()
        cam = OpenCamera(ListCameras()[cam_id])
        cam.settings.readoutSpeed=0 # 0=20MHz, 1=10MHz, 7=40MHz
        cam.settings.imageFormat = 'mono16'
        cam.settings.binning = self.binning
        cam.settings.emGain = gain
        cam.settings.triggerType = 0
        cam.settings.exposure = self.exposure - self.estimated_readout_lag
        cam.settings.blackoutMode=True
        cam.settings.Flush()
        cam.StartStreaming()
        frame = cam.GrabFrame()
        self.dtype = np.uint16
        buf = np.frombuffer(frame.stringBuffer,
                            dtype = self.dtype).reshape(
                                (frame.width,frame.height))
        self.h = buf.shape[1]
        self.w = buf.shape[0]
        self._init_variables(dtype = self.dtype)
        cam.StopStreaming()
        cam.CloseCamera()
        ReleaseDriver()
        display("[QImaging] Got info from camera (name: {0})".format(cam_id))
        self.camera_ready = Event()

    def run(self):

        ReleaseDriver()
        self.close_event.clear()
        while not self.close_event.is_set():
            self.nframes.value = 0
            LoadDriver()
            if not self.camera_ready.is_set():
                # prepare camera
                cam = OpenCamera(ListCameras()[self.cam_id])
                if cam.settings.coolerActive:
                    display('QImaging - cooler active.')
                cam.settings.readoutSpeed=0 # 0=20MHz, 1=10MHz, 7=40MHz
                cam.settings.imageFormat = 'mono16'
                cam.settings.binning = self.binning
                cam.settings.emGain = self.gain
                cam.settings.exposure = self.exposure - self.estimated_readout_lag
                if self.hardware_trigger.is_set():
                    triggerType = self.trigger_type
                else:
                    triggerType = 0
                cam.settings.triggerType = triggerType
                cam.settings.blackoutMode=True
                cam.settings.Flush()
                queue = CameraQueue(cam)
                display('QImaging - Camera ready!')
                self.camera_ready.set()
                self.nframes.value = 0
                # Wait for trigger
            while not self.start_trigger.is_set():
                # limits resolution to 1 ms 
                time.sleep(0.001)
                if self.close_event.is_set():
                    break
            if self.close_event.is_set():
                queue.stop()
                del queue
                del cam
                break
            queue.start()
            #tstart = time.time()
            display('QImaging - Started acquisition.')
            self.camera_ready.clear()
            while not self.stop_trigger.is_set():
                # run and acquire frames
                try:
                    f = queue.get(True, 1)
                except queue.Empty:
                    continue
                self.nframes.value += 1
                frame = np.ndarray(buffer = f.stringBuffer,
                                   dtype = self.dtype,
                                   shape = (self.w,
                                            self.h)).copy()
                    
                #display("Time {0} - {1}:".format(str(1./(time.time()-tstart)),self.nframes.value))
                #tstart = time.time()
                timestamp = f.timeStamp
                frameID = f.frameNumber
                if self.save_trigger.is_set():
                    self.was_saving = True
                    self.queue.put((frame.reshape([self.h,self.w]),
                                    (frameID,timestamp)))
                elif self.was_saving:
                    self.was_saving = False
                    self.queue.put(['STOP'])
                self.img[:] = np.reshape(frame,self.img.shape)[:]

                queue.put(f)

            queue.stop()
            del queue
            cam.settings.blackoutMode=False
            cam.settings.Flush()
            del cam
            self.save_trigger.clear()
            self.start_trigger.clear()
            self.stop_trigger.clear()
            ReleaseDriver()
            time.sleep(0.01)
            display('QImaging - Stopped acquisition.')



#  labcams - https://jpcouto@bitbucket.org/jpcouto/labcams.git
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .cams import *
# Allied Vision Technologies cameras
from pymba import *

def AVT_get_ids():
    with Vimba() as vimba:
        # get system object
        system = vimba.getSystem()
        # list available cameras (after enabling discovery for GigE cameras)
        if system.GeVTLIsPresent:
            system.runFeatureCommand("GeVDiscoveryAllOnce")
        #time.sleep(0.01)
        camsIds = vimba.getCameraIds()
        cams = [vimba.getCamera(id) for id in camsIds]
        camsModel = []
        for camid,cam in zip(camsIds,cams):
            try:
                cam.openCamera()
                
            except:
                camsModel.append('')
                continue
            camsModel.append('{0} {1} {2}'.format(cam.DeviceModelName,
                                                  cam.DevicePartNumber,
                                                  cam.DeviceID))
    return camsIds,camsModel

class AVTCam(GenericCam):    
    def __init__(self, cam_id = None, out_q = None,
                 start_trigger = None,
                 stop_trigger = None,
                 save_trigger = None,
                 exposure = 29000,
                 frame_rate = 30.,
                 gain = 10,
                 frame_timeout = 100,
                 n_frame_buffers = 10,
                 hardware_trigger = Event(),
                 trigger_source = 'Line1',
                 trigger_mode = 'LevelHigh',
                 trigger_selector = 'FrameStart',
                 acquisition_mode = 'Continuous',
                 n_triggered_frames = 1000,
                 frame_timeout = 100,
                 recorderpar = None):
        self.drivername = 'AVT'
        super(AVTCam,self).__init__(self,
                                    cam_id = cam_id,
                                    out_q = out_q,
                                    start_trigger = start_trigger,
                                    stop_trigger = stop_trigger,
                                    save_trigger = save_trigger,
                                    recorderpar=recorderpar)
        self.exposure = ((1000000/int(frameRate)) - 150)/1000.
        self.frame_rate = frameRate
        self.fs.value = self.frame_rate
        self.gain = gain
        self.frameTimeout = frameTimeout
        self.triggerSource = triggerSource
        self.triggerSelector = triggerSelector
        self.acquisitionMode = acquisitionMode
        self.nTriggeredFrames = nTriggeredFrames 
        self.nbuffers = nFrameBuffers
        self.frame_timeout = frame_timeout
        self.triggerMode = triggerMode
        self.tickfreq = float(1.0)
        with Vimba() as vimba:
            system = vimba.getSystem()
            if system.GeVTLIsPresent:
                system.runFeatureCommand("GeVDiscoveryAllOnce")
            time.sleep(0.01)
            self.cam = vimba.getCamera(cam_id)
            self.cam.openCamera()
            names = self.cam.getFeatureNames()
            # get a frame
            self.cam.acquisitionMode = 'SingleFrame'
            self.set_exposure(self.exposure)
            self.set_framerate(self.frame_rate)
            self.set_gain(self.gain)
            self.tickfreq = float(self.cam.GevTimestampTickFrequency)
            self.cam.TriggerSource = 'FixedRate'
            self.cam.TriggerMode = 'Off'
            self.cam.TriggerSelector = 'FrameStart'
            frame = self.cam.getFrame()
            frame.announceFrame()
            self.cam.startCapture()
            frame.queueFrameCapture()
            self.cam.runFeatureCommand('AcquisitionStart')
            frame.waitFrameCapture()
            self.cam.runFeatureCommand('AcquisitionStop')
            self.h = frame.height
            self.w = frame.width
            self.dtype = np.uint8
            self._init_variables(dtype = self.dtype)
            framedata = np.ndarray(buffer = frame.getBufferByteData(),
                                   dtype = self.dtype,
                                   shape = (frame.height,
                                            frame.width)).copy()
            self.img[:] = np.reshape(framedata,self.img.shape)[:]
            display("AVT [{1}] = Got info from camera (name: {0})".format(
                self.cam.DeviceModelName,self.cam_id))
            self.cam.endCapture()
            self.cam.revokeAllFrames()
            self.cam = None
        self.hardware_trigger = hardware_trigger
        if self.hardware_trigger is None:
            self.hardware_trigger = Event()
        if self.hardware_trigger.is_set():
            display('AVT [{0}] - Hardware triggered mode ON.'.format(self.cam_id))
            self.triggerSource = triggerSource
    def _init_controls(self):
        self.ctrevents = dict(
            exposure=dict(
                function = 'set_exposure',
                widget = 'float',
                variable = 'exposure',
                units = 'ms',
                type = 'float',
                min = 0.001,
                max = 100000,
                step = 10),
            gain = dict(
                function = 'set_gain',
                widget = 'float',
                variable = 'gain',
                units = 'ms',
                type = 'int',
                min = 0,
                max = 30,
                step = 1),
            framerate = dict(
                function = 'set_framerate',
                widget = 'float',
                type = 'float',
                variable = 'frame_rate',
                units = 'fps',
                min = 0.001,
                max = 1000,
                step = 1))
        
    def set_exposure(self,exposure = 30):
        '''Set the exposure time is in ms'''
        self.exposure = exposure
        if not self.cam is None:
            self.cam.ExposureTimeAbs =  int(self.exposure*1000)
            display('[AVT {0}] Setting exposure to {1} ms.'.format(
                self.cam_id, self.exposure))

    def set_framerate(self,frame_rate = 30):
        '''set the frame rate of the AVT camera.''' 
        self.frame_rate = frame_rate
        if not self.cam is None:
            self.cam.AcquisitionFrameRateAbs = self.frame_rate
            self.fs.value = self.frame_rate
            if self.cam_is_running:
                self.start_trigger.set()
                self.stop_trigger.set()
            display('[AVT {0}] Setting frame rate to {1} .'.format(
                self.cam_id, self.frame_rate))
    def set_gain(self,gain = 0):
        ''' Set the gain of the AVT camera'''
        self.gain = int(gain)
        if not self.cam is None:
            self.cam.GainRaw = self.gain
            display('[AVT {0}] Setting camera gain to {1} .'.format(
                self.cam_id, self.gain))
    
    def _cam_init(self):
        self.nframes.value = 0
        self.recorded_frames = []
        self.vimba = Vimba()
        self.vimba.startup()
        system = self.vimba.getSystem()
        if system.GeVTLIsPresent:
            system.runFeatureCommand("GeVDiscoveryAllOnce")
            time.sleep(0.1)
        # prepare camera
        self.cam = self.vimba.getCamera(self.cam_id)
        self.cam.openCamera()
        # cam.EventSelector = 'FrameTrigger'
        self.cam.EventNotification = 'On'
        self.cam.PixelFormat = 'Mono8'
        self.cameraFeatureNames = self.cam.getFeatureNames()
        #display('\n'.join(cameraFeatureNames))
        self.set_framerate(self.frame_rate)
        self.set_gain(self.gain)
        self.set_exposure(self.exposure)
        
        self.cam.SyncOutSelector = 'SyncOut1'
        self.cam.SyncOutSource = 'FrameReadout'#'Exposing'
        if self.hardware_trigger.is_set():
            self.cam.TriggerSource = self.triggerSource#'Line1'#self.triggerSource
            self.cam.TriggerMode = 'On'
            #cam.TriggerOverlap = 'Off'
            self.cam.TriggerActivation = self.triggerMode #'LevelHigh'##'RisingEdge'
            self.cam.AcquisitionMode = self.acquisitionMode
            self.cam.TriggerSelector = self.triggerSelector
            if self.acquisitionMode == 'MultiFrame':
                self.cam.AcquisitionFrameCount = self.nTriggeredFrames
                self.cam.TriggerActivation = self.triggerMode #'LevelHigh'##'RisingEdge'
        else:
            display('[Cam - {0}] Using no trigger.'.format(self.cam_id))
            self.cam.AcquisitionMode = 'Continuous'
            self.cam.TriggerSource = 'FixedRate'
            self.cam.TriggerMode = 'Off'
            self.cam.TriggerSelector = 'FrameStart'
        # create new frames for the camera
        self.frames = []
        for i in range(self.nbuffers):
            self.frames.append(self.cam.getFrame())    # creates a frame
            self.frames[i].announceFrame()
        self.cam.startCapture()
        for f,ff in enumerate(self.frames):
            try:
                ff.queueFrameCapture()
            except:
                display('Queue frame error while getting cam ready: '+ str(f))
                continue                    
        self.camera_ready.set()
        self.nframes.value = 0
        # Ready to wait for trigger
            
    def _cam_startacquisition(self):
        self.cam.runFeatureCommand("GevTimestampControlReset")
        self.cam.runFeatureCommand('AcquisitionStart')
        if self.hardware_trigger.is_set():
            self.cam.TriggerSelector = self.triggerSelector
            self.cam.TriggerMode = 'On'
        #tstart = time.time()
        self.lastframeid = [-1 for i in self.frames]

    def _cam_loop(self):
        # run and acquire frames
        #sortedfids = np.argsort([f._frame.frameID for f in frames])
        for ibuf in range(self.nbuffers):
            f = self.frames[ibuf]
            avterr = f.waitFrameCapture(timeout = self.frameTimeout)
            if avterr == 0:
                timestamp = f._frame.timestamp/self.tickfreq
                frameID = f._frame.frameID
                #print('Frame id:{0}'.format(frameID))
                if not frameID in self.recorded_frames:
                    self.recorded_frames.append(frameID)
                    frame = np.ndarray(buffer = f.getBufferByteData(),
                                       dtype = self.dtype,
                                       shape = (f.height,
                                                f.width)).copy()
                    #display("Time {0} - {1}:".format(str(1./(time.time()-tstart)),self.nframes.value))
                    #tstart = time.time()
                    try:
                        f.queueFrameCapture()
                    except:
                        display('Queue frame failed: '+ str(f))
                        return None,(None,None)
                    self.lastframeid[ibuf] = frameID
                    return frame,(frameID,timestamp)
            elif avterr == -12:
                #display('VimbaException: ' +  str(avterr))        
                return None,(None,None)

    def _cam_close(self):
        self.cam.runFeatureCommand('AcquisitionStop')
        display('[AVT] - Stopped acquisition.')
        # Check if all frames are done...
        for ibuf in range(self.nbuffers):
            f = self.frames[ibuf]
            try:
                f.waitFrameCapture(timeout = self.frame_timeout)
                timestamp = f._frame.timestamp/self.tickfreq
                frameID = f._frame.frameID
                frame = np.ndarray(buffer = f.getBufferByteData(),
                                   dtype = self.dtype,
                                   shape = (f.height,
                                            f.width)).copy()
                if self.save_trigger.is_set():
                    self.was_saving = True
                    if not frameID in self.lastframeid :
                        self.queue.put((frame.copy(),(frameID,timestamp)))
                elif self.was_saving:
                    self.was_saving = False
                    self.queue.put(['STOP'])

                self.lastframeid[ibuf] = frameID
                self.nframes.value = frameID
                self.frame = frame
            except VimbaException as err:
                #display('VimbaException: ' + str(err))
                pass
        display('{4} delivered:{0},dropped:{1},queued:{4},time:{2}'.format(
            self.cam.StatFrameDelivered,
            self.cam.StatFrameDropped,
            self.cam.StatTimeElapsed,
            self.cam.DeviceModelName,
            self.nframes.value))
        self.cam.runFeatureCommand('AcquisitionStop')
        self.cam.endCapture()
        try:
            self.cam.revokeAllFrames()
        except:
            display('Failed to revoke frames.')
        self.cam.closeCamera()
        display('AVT [{0}] - Close event: {1}'.format(
            self.cam_id,
            self.close_event.is_set()))
        self.vimba.shutdown()



#  labcams - https://jpcouto@bitbucket.org/jpcouto/labcams.git
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .utils import *
from .cams import * 
from .io import *
from .widgets import *

LOGO = '''
                                             MMM
                                           MMMMMM
    MMM:               .MMMM             MMMM MMMMMMMM
    MMM:               .MMMM            MMMMM MMMMMMMM      
    MMM:               .MMMM             MMMM  MMMMMM        MM 
    MMM:  :MMMMMMMMM.  .MMMMOMMMMMM       MN     MMM      MMMMM 
    MMM:  :M     MMMM  .MMMMM?+MMMMM    MMMMMMMMMMMMMMM7MMMMMMM  
    MMM:         OMMM  .MMMM    MMMM    MMMMMMMMMMMMMMMMMMMMMMM 
    MMM:  .MMMMMMMMMM  .MMMM    ?MMM    MMMMMMMMMMMMMMMMMMMMMMM 
    MMM:  MMMM  .8MMM  .MMMM    ZMMM    MMMMMMMMMMMMMMMMMMMMMMM 
    MMM:  MMM=...8MMM  .MMMM    MMMM    MMMMMMMMMMMMMMM.MMMMMMM  
    MMM:  MMMMMMMMMMM  .MMMMMMMMMMM                        MMMM 
    MMM:   MMMMM 8MMM  .MMMM:MMMMZ                            M 

         MMMMMMN  =MMMMMMMM     MMMM.MMMM$ .+MMMM      MMMMMMM: 
       MMMMMMMMM  +MMMMMMMMM$   MMMMMMMMMMMMMMMMMM   MMMMMMMMM8 
      MMMM               MMMM   MMMM   MMMM    MMM+  MMM8       
      MMMZ          OMMMMMMMM   MMMM   NMMM    MMM?  MMMMMMM$   
      MMMI        MMMMM  MMMM   MMMM   NMMM    MMM?   ZMMMMMMMM 
      MMMM       7MMM    MMMM   MMMM   NMMM    MMM?        MMMM 
       MMMMD+7MM  MMMN   MMMM   MMMM   NMMM    MMM?  MM$:.7MMMM   
        MMMMMMMM  ZMMMMMOMMMM   MMMM   NMMM    MMM?  MMMMMMMM+                                                              
                          https://bitbucket.org/jpcouto/labcams
'''

N_UDP = 1024

class LabCamsGUI(QMainWindow):
    app = None
    cams = []
    def __init__(self,app = None, expName = 'test',
                 camDescriptions = [],
                 parameters = {},
                 server = False,
                 save_on_start = False,
                 hardware_trigger = False,
                 software_trigger = True,
                 update_frequency = 33):
        '''
        Graphical interface for controling labcams.
        '''
        super(LabCamsGUI,self).__init__()

        self.software_trigger = software_trigger # one trigger per camera so we can control the timing (daq sleeps before)
        self.hardware_trigger_event = Event()
        if hardware_trigger:
            self.hardware_trigger_event.set()
        else:
            self.hardware_trigger_event.clear()

        self.parameters = parameters
        self.app = app
        self.plugins = []
        self.update_frequency = update_frequency
        self.save_on_start = save_on_start
        self.cam_descriptions = camDescriptions
        self.zmqsocket = None
        self.udpsocket = None
        if server:
            if not 'server_refresh_time' in self.parameters.keys():
                self.parameters['server_refresh_time'] = 5
            if not 'server' in self.parameters.keys():
                self.parameters['server'] = 'zmq'
            if self.parameters['server'] == 'udp':
                import socket
                self.udpsocket = socket.socket(socket.AF_INET, 
                                     socket.SOCK_DGRAM) # UDP
                self.udpsocket.bind(('0.0.0.0',
                                     self.parameters['server_port']))
                display('Listening to UDP port: {0}'.format(
                    self.parameters['server_port']))
                self.udpsocket.settimeout(.02)
            else:
                import zmq
                self.zmqContext = zmq.Context()
                self.zmqsocket = self.zmqContext.socket(zmq.REP)
                self.zmqsocket.bind('tcp://0.0.0.0:{0}'.format(
                    self.parameters['server_port']))
                display('Listening to ZMQ port: {0}'.format(
                    self.parameters['server_port']))
            self.server_timer = QTimer()
            self.server_timer.timeout.connect(self.server_actions)
            self.server_timer.start(self.parameters['server_refresh_time'])
        
        # Init cameras
        if not 'recorder_path_format' in self.parameters.keys():
            print('Using default recorder_path_format value.')
            self.parameters['recorder_path_format'] = pjoin('{datafolder}',
                                                            '{dataname}',
                                                            '{filename}',
                                                            '{today}_{run}_{nfiles}')
        if not 'recorder_path' in self.parameters.keys():
            self.parameters['recorder_path'] = pjoin(os.path.expanduser('~'),'data'),

        if not 'recorder_frames_per_file' in self.parameters.keys():
            self.parameters['recorder_frames_per_file'] = 0

        camdrivers = [cam['driver'].lower() for cam in camDescriptions]
        if 'avt' in camdrivers:
            from .avt import AVT_get_ids
            try:
                avtids,avtnames = AVT_get_ids()
            except Exception as e:
                display('[ERROR] AVT  camera error? Connections? Parameters?')
                print(e)
        self.camQueues = []
        self.saveflags = []
        connected_avt_cams = []
        for c,cam in enumerate(self.cam_descriptions):
            display("Connecting to camera [" + str(c) + '] : '+cam['name'])
            if not 'save_data' in cam.keys():   # to disable the saving for an individual camera
                cam['save_data'] = True
            self.saveflags.append(cam['save_data'])                
            cam = dict(cam,
                       filename = expName,
                       recorder_path = self.parameters['recorder_path'],
                       recorder_path_format = self.parameters['recorder_path_format'])
            # default recorder
            if not 'recorder' in cam.keys():
                cam['recorder'] = dict(format='tiff',  
                                       method = 'queue',
                                       compression = 0,
                                       frames_per_file = 1024,
                                       sleep_time = 0.3)
            self.cams.append(Camera(**cam,
                                    hardware_trigger_event = self.hardware_trigger_event))
            if hasattr(self.cams[-1],'excitation_trigger'):
                self.excitation_trigger = self.cams[-1].excitation_trigger
                self.excitation_trigger_widget = CamStimTriggerWidget(
                    ino = self.cams[-1].excitation_trigger,
                    cam = self.cams[-1].cam)
                self.camstim_tab = QDockWidget("Camera excitation control",self)
                self.camstim_tab.setWidget(self.excitation_trigger_widget)
                self.addDockWidget(
                    Qt.LeftDockWidgetArea,
                    self.camstim_tab)
            # Print parameters
            display('\t Camera: {0}'.format(cam['name']))
            for k in np.sort(list(cam.keys())):
                if not k == 'name' and not k == 'recorder':
                    display('\t\t - {0} {1}'.format(k,cam[k]))
        self.initUI()
        
        self.camerasRunning = False
        
        for cam in self.cams[::-1]:
            cam.start()
        
        camready = 0
        while camready != len(self.cams):
            camready = np.sum([cam.camera_ready.is_set() for cam in self.cams])
        display('[labcams] - Initialized cameras.')
        
        self.recController.saveOnStartToggle.setChecked(self.save_on_start)
        self.recController.softTriggerToggle.setChecked(self.software_trigger)

    def set_experiment_name(self,expname):
        # Makes sure that the experiment name has the right slashes.
        if os.path.sep == '/':
            expname = expname.replace('\\',os.path.sep)
        expname = expname.strip(' ')
        for flg,cam in zip(self.saveflags,self.cams):
            if flg:
                cam.set_filename(expname)
        self.recController.experimentNameEdit.setText(expname)

    def server_reply(self,msg, msgtype = 'ok',address = None):
        if not self.zmqsocket is None:
            self.zmqsocket.send_pyobj(dict(action=msgtype,
                                           value = msg))
        if not self.udpsocket is None and not address is None:
            self.udpsocket.sendto('{0}={1}'.format(msgtype,msg).encode(),address)
                
    def server_actions(self): # all this should be moved to a class somewhere else.
        if not self.zmqsocket is None:
            try:
                message = self.zmqsocket.recv_pyobj(flags=zmq.NOBLOCK)
            except:
                return
        elif self.parameters['server'] == 'udp':
            try:
                msg,address = self.udpsocket.recvfrom(N_UDP)
            except:
                return
            msg = msg.decode().split('=')
            message = dict(action=msg[0])
            if len(msg) > 1:
                message = dict(message,value=msg[1])
            display('Server received message: {0}'.format(message))            
        if message['action'].lower() == 'expname':
            self.set_experiment_name(message['value'])
            self.server_reply(msg = message['action'].lower(),address = address)
        elif message['action'].lower() in ['softtrigger','software_trigger','settrigger','trigger']:
            self.recController.softTriggerToggle.setChecked(
                int(message['value']))
            self.server_reply(msg = 'trigger',address = address)
        elif message['action'].lower() == ['hardtrigger','hardware_trigger']:
            self.recController.camTriggerToggle.setChecked(
                int(message['value']))
            self.server_reply(msg = 'save_hardwaretrigger',address = address) 
        elif message['action'].lower() in ['save','setmanualsave','manualsave','acquire']:
            self.recController.saveOnStartToggle.setChecked(
                int(message['value']))
            self.server_reply(msg = 'save',address = address) 
        elif message['action'].lower() == 'log':
            for c in self.cams:
                c.cam.eventsQ.put('log={0}'.format(message['value']))
            self.recController.udpmessages.setText(message['value'])
            self.server_reply(msg = 'log',address = address) 
        elif message['action'].lower() == 'snapshot':
            foldername = message['value']
            if not os.path.exists(foldername):
                os.path.makedirs(foldername)
                display('Created {0}'.format(foldername))
            display('Getting snapshots to {0}'.format(foldername))
            update_shared_date()
            for icam,cam in enumerate(self.cams):
                frame = cam.get_img_with_virtual_channels()
                dataname = cam.recorder_parameters['dataname']
                fname = pjoin(os.path.dirname(foldername),'snapshots',
                              shared_date[:]+'_{0}.tif'.format(dataname))
                from tifffile import imsave
                imsave(fname,
                       frame.transpose([2,0,1]).squeeze(),
                       metadata = {
                           'Camera':str(icam)})
            self.server_reply(msg = 'snapshots',address = address) 
        elif message['action'].lower() == 'load_reference':
            foldername = message['value']
            if not os.path.exists(foldername):
                display('No folder: {0}'.format(foldername))
                self.server_reply(msg = 'no_folder',msgtype = 'error',address = address) 
                return
            for icam,cam in enumerate(self.cams):
                files = glob(pjoin(foldername,'*_{0}.tif').format(
                    cam.recorder_parameters['dataname']))
                if len(files):
                    self.camwidgets[icam].toggle_reference(filename = files[0])
            self.server_reply(msg = 'load_reference',address = address)
        elif message['action'].lower() == 'hide_reference':
            for icam,cam in enumerate(self.cams):
                self.camwidgets[icam].toggle_reference(filename = '')
            self.server_reply(msg = 'hide_reference',address = address) 
        elif message['action'].lower() == 'ping':
            display('Server got PING.')
            self.server_reply(msg = 'pong',address = address) 
        elif message['action'].lower() == 'quit':
            self.udpsocket.sendto(b'ok=bye',address)
            self.server_reply(msg = 'bye',address = address) 
            self.close()
        
    def experiment_menu_trigger(self,q):
        if q.text() == 'Set refresh time':
            self.timer.stop()
            res = QInputDialog().getDouble(self,"What refresh period do you want?","GUI refresh period",
                                           self.update_frequency)
            if res[1]:
                self.update_frequency = res[0]
            self.timer.start(self.update_frequency)
            #display(q.text()+ "clicked. ")
        
    def initUI(self):
        # Menu
        self.setDockOptions(QMainWindow.AllowTabbedDocks |
                            QMainWindow.AllowNestedDocks
)
        bar = self.menuBar()
        editmenu = bar.addMenu("Options")
        editmenu.addAction("Set refresh time")
        editmenu.triggered[QAction].connect(self.experiment_menu_trigger)
        self.setWindowTitle("labcams")
        self.tabs = []
        self.camwidgets = []
        self.recController = RecordingControlWidget(self)
        #self.setCentralWidget(self.recController)
        self.recControllerTab = QDockWidget("",self)
        self.recControllerTab.setWidget(self.recController)
        self.addDockWidget(
            Qt.TopDockWidgetArea,
            self.recControllerTab)
        self.recController.setFixedHeight(self.recController.layout.sizeHint().height())
        for c,cam in enumerate(self.cams):
            tt = ''
            if self.saveflags[c]:
                tt +=  ' - ' + self.cam_descriptions[c]['name'] +' ' 
            self.tabs.append(QDockWidget("Camera: "+str(c) + tt,self))
            if hasattr(cam.cam,"h") and hasattr(cam.cam,"w"): # then it must be a camera
                self.camwidgets.append(CamWidget(frame = np.zeros((cam.cam.h.value,
                                                                   cam.cam.w.value,
                                                                   cam.cam.nchan.value),
                                                                  dtype=cam.cam.dtype),
                                                 iCam = c,
                                                 parent = self,
                                                 parameters = self.cam_descriptions[c]))
                self.camwidgets[-1].setMinimumHeight(300)

            else: # NIDAQ is the only other camera so add that widget
                self.camwidgets.append(DAQPlotWidget(daq = cam.cam,
                                                     parent = self,
                                                     parameters = self.cam_descriptions[c]))
            self.tabs[-1].setWidget(self.camwidgets[-1])
            self.tabs[-1].setFloating(False)
            self.tabs[-1].setAllowedAreas(Qt.LeftDockWidgetArea |
                                          Qt.RightDockWidgetArea |
                                          Qt.BottomDockWidgetArea |
                                          Qt.TopDockWidgetArea)
            self.tabs[-1].setFeatures(QDockWidget.DockWidgetMovable |
                                      QDockWidget.DockWidgetFloatable)
            self.addDockWidget(
                Qt.BottomDockWidgetArea,
                self.tabs[-1])
            
            # there can only be one of these for now?
            if hasattr(self,'camstim_widget'):
                self.camstim_tab = QDockWidget("Camera excitation control",self)
                self.camstim_tab.setWidget(self.camstim_widget)
                self.addDockWidget(
                    Qt.LeftDockWidgetArea,
                self.camstim_tab)
            display('Init view: ' + str(c))
        self.timer = QTimer()
        self.timer.timeout.connect(self.update_timer)
        self.timer.start(self.update_frequency)
        #self.move(0, 0)
        self.show()
            	
    def update_timer(self):
        for c,cam in enumerate(self.cams):
            self.camwidgets[c].update()
            try:
                pass
            except Exception as e:
                display('Could not draw cam: {0}'.format(c))
                exc_type, exc_obj, exc_tb = sys.exc_info()
                fname = os.path.split(
                    exc_tb.tb_frame.f_code.co_filename)[1]
                print(e, fname, exc_tb.tb_lineno)
        for p in self.plugins:  # update plugins
            try:
                p.update()
            except Exception as err:
                display('Could not update plugin')
                exc_type, exc_obj, exc_tb = sys.exc_info()
                fname = os.path.split(
                    exc_tb.tb_frame.f_code.co_filename)[1]
                print(e, fname, exc_tb.tb_lineno)
            
    def closeEvent(self,event):
        if hasattr(self,'server_timer'):
            self.server_timer.stop()
            if hasattr(self,'udpsocket') and not self.udpsocket is None:
                self.udpsocket.close()
        self.timer.stop()            
        display('Acquisition stopped (close event).')
        for cam in self.cams:
            cam.stop_acquisition()
        for cam in self.cams:
            cam.close()
        pg.setConfigOption('crashWarning', False)
        event.accept()


def main():

    from argparse import ArgumentParser, RawDescriptionHelpFormatter
    import os
    import json
    parser = ArgumentParser(description=LOGO + '\n\n  Multiple camera control and recording.',formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('file',
                        metavar='file',
                        type=str,
                        default=None,
                        nargs="?")
    parser.add_argument('-d','--make-config',
                        type=str,
                        default = None,
                        action='store')
    parser.add_argument('-w','--wait',
                        default = False,
                        action='store_true')
    parser.add_argument('-t','--triggered',
                        default=False,
                        action='store_true')
    parser.add_argument('-c','--cam-select',
                        type=int,
                        nargs='+',
                        action='store')
    parser.add_argument('--no-server',
                        default=False,
                        action='store_true')
    parser.add_argument('--bin-to-mj2',
                        default=False,
                        action='store_true')
    parser.add_argument('--mj2-rate',
                        default=30.,
                        action='store')
    
    opts = parser.parse_args()

    if opts.bin_to_mj2:
        from labcams.io import mmap_dat
        
        fname = opts.file
        
        assert not fname is None, "Need to supply a binary filename to compress."
        assert os.path.isfile(fname), "File {0} not found".format(fname)
        ext = os.path.splitext(fname)[-1]
        assert ext in ['.dat','.bin'], "File {0} needs to be binary.".format(fname)  
        stack = mmap_dat(fname)
        stack_to_mj2_lossless(stack, fname, rate = opts.mj2_rate)
        print('Converted {0}'.format(fname.replace(ext,'.mov')))
        sys.exit(0)
        
    if not opts.make_config is None:
        from .widgets import SettingsDialog
        app = QApplication(sys.argv)
        s = SettingsDialog(getPreferences())
        sys.exit(app.exec_())
        fname = opts.make_config
        getPreferences(fname, create=True)
        sys.exit(s.exec_())
    parameters = getPreferences(opts.file)
    cams = parameters['cams']
    if not opts.cam_select is None:
        cams = [parameters['cams'][i] for i in opts.cam_select]

    app = QApplication(sys.argv)
    w = LabCamsGUI(app = app,
                   camDescriptions = cams,
                   parameters = parameters,
                   server = not opts.no_server,
                   software_trigger = not opts.wait,
                   hardware_trigger = opts.triggered)
    sys.exit(app.exec_())
    
if __name__ == '__main__':
    main()



#  labcams - https://jpcouto@bitbucket.org/jpcouto/labcams.git
# Copyright (C) 2020 Joao Couto - jpcouto@gmail.com
#
#  This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Classes to save files from a multiprocessing queue
import time
import sys
from multiprocessing import Process,Queue,Event,Array,Value
from multiprocessing.shared_memory import SharedMemory # this breaks compatibility with python < 3.8
from ctypes import c_long, c_char_p, c_wchar
import ctypes
from datetime import datetime
import time
import sys
from .utils import display,shared_date
import numpy as np
import os
from glob import glob
from os.path import join as pjoin
from tifffile import imread, TiffFile
from tifffile import TiffWriter as twriter
import pandas as pd
from skvideo.io import FFmpegWriter
import cv2

# TODO: check if ffmpeg is working when initializing and using the ffmpeg writer.
VERSION = '0.7'

class GenericWriter(object):
    def __init__(self,
                 cam = None,
                 loggerQ = None,
                 filename = pjoin('dummy','run'),
                 dataname = 'eyecam',
                 pathformat = pjoin('{datafolder}','{dataname}','{filename}',
                                    '{today}_{run}_{nfiles}'),
                 datafolder=pjoin(os.path.expanduser('~'),'data'),
                 framesperfile=0,
                 sleeptime = 1./30,
                 virtual_channels = None,
                 incrementruns=True,
                 **kwargs):
        if not hasattr(self,'extension'):
            self.extension = '.nan'
        self.cam = None
        self.saved_frame_count = 0
        self.runs = 0
        self.write = False
        self.close = False
        self.sleeptime = sleeptime # seconds
        self.framesperfile = framesperfile
        self.filename = ''
        self.datafolder = datafolder
        self.dataname = dataname
        self.foldername = None
        self.incrementruns = incrementruns
        self.fd = None
        self.parQ = None
        self.today = datetime.today().strftime('%Y%m%d')
        self.nchannels = Value('i',1)
        self.logfile = None
        self.nFiles = 0
        runname = 'run{0:03d}'.format(self.runs)
        self.virtual_channels = virtual_channels
        if self.virtual_channels is None:
            self.virtual_channels = Value('i',0)

        self.path_format = pathformat
        self.path_keys =  dict(datafolder = self.datafolder,
                               dataname = self.dataname,
                               filename = self.filename,
                               today = self.today,
                               datetime = shared_date[:],
                               year = shared_date[:4],
                               month = shared_date[4:6],
                               day = shared_date[6:8],
                               hours = shared_date[9:11],
                               minutes = shared_date[11:13],
                               seconds = shared_date[13:],
                               run = runname,
                               nfiles = '{0:08d}'.format(0),
                               extension = self.extension,
                               **kwargs)
        if self.framesperfile > 0:
            if not '{nfiles}' in self.path_format:
                self.path_format += '_{nfiles}'
        if not cam is None:
            self.init_cam(cam)
            
    def init_cam(self,cam):
        ''' Sets camera specific variables - happens after camera load'''
        if not hasattr(cam,'membuffer_name'):
            return
        self.frame_rate = None
        if self.cam is None:
            self.cam = dict(buffer_name = cam.membuffer_name,
                            buffer_len = cam.membuffer_len,
                            queue = cam.queue,
                            dtype = cam.dtype,
                            h = cam.h,
                            w = cam.w,
                            nchannels = cam.nchan,
                            frame_rate = cam.fs)
            self.inQ = self.cam['queue']
        if hasattr(cam,'frame_rate'):
            self.frame_rate = cam.fs.value
        self.nchannels = self.cam['nchannels']
        self.h = self.cam['h']
        self.w = self.cam['w']
        if self.virtual_channels.value == 0:    # to set the filename in the binary file for instance.
            self.virtual_channels.value = self.nchannels.value
        
    def _stop_write(self):
        self.write = False
    def stop(self):
        self.write = False
    def set_filename(self,filename):
        self._stop_write()
        self.filename = filename
        display('Filename updated: ' + self.get_filename())

    def get_filename(self):
        return str(self.filename[:]).strip(' ')

    def get_filename_path(self):
        self.path_keys['run'] = 'run{0:03d}'.format(self.runs)
        nfiles = self.nFiles
        self.path_keys['nfiles'] = '{0:08d}'.format(nfiles)
        self.path_keys['datetime'] = shared_date[:]
        self.path_keys['year'] = shared_date[:4]
        self.path_keys['month'] = shared_date[4:6]
        self.path_keys['day'] = shared_date[6:8]
        self.path_keys['hours'] = shared_date[9:11]
        self.path_keys['minutes'] = shared_date[11:13]
        self.path_keys['seconds'] = shared_date[13:]
        
        self.path_keys['filename'] = self.get_filename()
        filename = (self.path_format+'{extension}').format(**self.path_keys)
        folder = os.path.dirname(filename)
        if folder == '':
            filename = pjoin(os.path.abspath(os.path.curdir),filename)
            folder = os.path.dirname(filename)
        if not os.path.exists(folder):
            try:
                os.makedirs(folder)
            except Exception as e:
                print("Could not create folder {0}".format(folder))
                print(e)
        return filename

    def open_file(self,nfiles = None,frame = None):
        filename = self.get_filename_path()
        if not self.fd is None:
            self.close_file()
        self._open_file(filename,frame)
        # Create a log file
        if self.logfile is None:
            self._open_logfile()
        self.nFiles += 1
        if hasattr(self,'parsed_filename'):
            filename = self.parsed_filename
        display('Opened: '+ filename)        
        self.logfile.write('# [' + datetime.today().strftime('%y-%m-%d %H:%M:%S')+'] - ' + filename + '\n')

    def _open_logfile(self):
        #self.path_keys['run'] = 'run{0:03d}'.format(self.runs)
        #nfiles = self.nFiles
        #self.path_keys['nfiles'] = '{0:08d}'.format(nfiles)
        #self.path_keys['filename'] = self.get_filename()

        #filename = (self.path_format+'{extension}').format(**self.path_keys)
        filename = self.get_filename_path()
        logfname = filename.replace('{extension}'.format(
            **self.path_keys),'.camlog')
        
        self.logfile = open(logfname,'w',encoding = 'utf-8')
        self.logfile.write('# Camera: {0} log file'.format(
            self.dataname) + '\n')
        self.logfile.write('# Date: {0}'.format(
            datetime.today().strftime('%d-%m-%Y')) + '\n')
        self.logfile.write('# labcams version: {0}'.format(
            VERSION) + '\n')                
        self.logfile.write('# Log header:' + 'frame_id,timestamp' + '\n')

    def _open_file(self,filename,frame):
        pass

    def _write(self,frame,frameid,timestamp):
        pass

    def save(self,frame,metadata):
        return self._handle_frame((frame,metadata))
    
    def _handle_frame(self,buff):
        if buff[0] is None:
            # Then parameters were passed to the queue
            display('[Recorder] - Received None...')
            return None,None
        if len(buff) == 1:
           # check message:
            msg = buff[0]
            if msg in ['STOP']:
                display('[Recorder] Stopping the recorder.')
                self._stop_write()
            elif msg.startswith('#'):
                if self.logfile is None:
                    self._open_logfile()
                self.logfile.write(msg + '\n')
            return None,msg
        else:
            frame,(metadata) = buff
            if (self.fd is None or
                (self.framesperfile > 0 and np.mod(self.saved_frame_count,
                                                   self.framesperfile)==0)):
                self.open_file(frame = frame)
                if not self.inQ is None:
                    display('Queue size: {0}'.format(self.inQ.qsize()))

                    self.logfile.write('# [' + datetime.today().strftime('%y-%m-%d %H:%M:%S')+'] - '
                                       + 'Queue: {0}'.format(self.inQ.qsize())
                                       + '\n')
            frameid, timestamp = metadata[:2] 
            self._write(frame,frameid,timestamp)
            if np.mod(frameid,7000) == 0:
                if self.inQ is None:
                    display('[{0} - frame:{1}]'.format(
                        self.dataname,frameid))
                else:
                    display('[{0} - frame:{1}] Queue size: {2}'.format(
                        self.dataname,frameid,self.inQ.qsize()))
            self.logfile.write(','.join(['{0}'.format(a) for a in metadata]) + '\n')
            self.saved_frame_count += 1
        return frameid,frame
    
    def close_run(self):
        
        if not self.logfile is None:
            # Check if there are comments on the queue
            while not self.inQ.empty():
                buff = self.inQ.get()
                frameid,frame = self._handle_frame(buff)
            self.close_file()
            self.logfile.write('# [' +
                               datetime.today().strftime(
                                   '%y-%m-%d %H:%M:%S')+'] - ' +
                               "Wrote {0} frames on {1} ({2} files).".format(
                                   self.saved_frame_count,
                                   self.dataname,
                                   self.nFiles) + '\n')
            self.logfile.close()
            self.logfile = None
            display('[Recorder] Closing the logfile {0}.'.format(self.dataname))
            self.runs += 1
        if not self.saved_frame_count == 0:
            display("[Recorder] Wrote {0} frames on {1} ({2} files).".format(
                self.saved_frame_count,
                self.dataname,
                self.nFiles))

class GenericWriterProcess(Process,GenericWriter):
    def __init__(self,
                 cam = None,
                 loggerQ = None,
                 filename = pjoin('dummy','run'),
                 dataname = 'eyecam',
                 pathformat = pjoin('{datafolder}','{dataname}','{filename}',
                                    '{today}_{run}_{nfiles}'),
                 datafolder=pjoin(os.path.expanduser('~'),'data'),
                 framesperfile=0,
                 sleeptime = 1./30,
                 virtual_channels = None,
                 incrementruns=True,
                 save_trigger = None,
                 **kwargs):
        GenericWriter.__init__(self,
                               cam = cam,
                               loggerQ=loggerQ,
                               filename=filename,
                               datafolder=datafolder,
                               dataname=dataname,
                               virtual_channels = virtual_channels,
                               pathformat=pathformat,
                               framesperfile=framesperfile,
                               sleeptime=sleeptime,
                               incrementruns=incrementruns,
                               **kwargs)
        Process.__init__(self)
        self.write = save_trigger
        if self.write is None:
            self.write = Event()
        self.close = Event()
        self.filename = Array('u',' ' * 1024)
        self.parQ = Queue()
        self.daemon = True

    def _stop_write(self):
        self.write.clear()

    def set_filename(self,filename):
        self._stop_write()
        for i in range(len(self.filename)):
            self.filename[i] = ' '
        for i in range(len(filename)):
            self.filename[i] = filename[i]
        display('Filename updated: ' + self.get_filename())
    
    def stop(self):
        self._stop_write()
        self.close.set()
        self.join()
        
    def _write(self,frame,frameid,timestamp):
        pass
    
    def get_from_queue_and_save(self):
        buff = self.inQ.get()
        buf = None
        if not buff[0] is None:
            if len(buff) > 1:
                buf = self.get_frame(buff[0])
                buff = [buf, *buff[1:]]
        return self._handle_frame(buff)

    def _init_shared_mem(self):
        dtype = self.cam['dtype']
        try:
            dtype = dtype()
        except:
            pass
        if not hasattr(self,'membuffer'):
            self.membuffer = SharedMemory(name = self.cam['buffer_name'])
        buffsize = [self.h.value,self.w.value,self.nchannels.value]
        self.nbuffers = int(self.cam['buffer_len'] // np.prod(buffsize+[dtype.itemsize]))
        buffsize = [self.nbuffers] + buffsize
        self.imgs = np.ndarray(buffsize,
                               buffer = self.membuffer.buf,
                               dtype = dtype)

    def get_frame(self,frame_index = None):
        if frame_index is None:
            frame_index = self.nframes.value
        return self.imgs[frame_index % self.nbuffers].squeeze()
        
    def run(self):
        while not self.close.is_set():
            self.saved_frame_count = 0
            self.nFiles = 0
            if not self.parQ.empty():
                self.getFromParQueue()
            self._init_shared_mem()
            while self.write.is_set() and not self.close.is_set():
                while self.inQ.qsize() > 0:
                    frameid,frame = self.get_from_queue_and_save()
                # spare the processor just in case...
                time.sleep(self.sleeptime)
            time.sleep(self.sleeptime)
            # If queue is not empty, empty if to disk.
            while self.inQ.qsize() > 0:
                frameid,frame = self.get_from_queue_and_save()
            #display('Queue is empty. Proceding with close.')
            # close the run
            self.close_run()
        
class TiffWriter(GenericWriterProcess):
    def __init__(self,
                 cam,
                 loggerQ = None,
                 filename = pjoin('dummy','run'),
                 dataname = 'cam',
                 pathformat = pjoin('{datafolder}','{dataname}','{filename}',
                                    '{today}_{run}_{nfiles}'),
                 datafolder=pjoin(os.path.expanduser('~'),'data'),
                 framesperfile=256,
                 sleeptime = 1./30,
                 incrementruns=True,
                 compression=None,
                 **kwargs):
        self.extension = '.tif'
        super(TiffWriter,self).__init__(cam = cam,
                                        loggerQ=loggerQ,
                                        datafolder=datafolder,
                                        filename=filename,
                                        dataname=dataname,
                                        pathformat=pathformat,
                                        framesperfile=framesperfile,
                                        sleeptime=sleeptime,
                                        incrementruns=incrementruns)
        self.compression = None
        if not compression is None:
            if compression > 9:
                display('Can not use compression over 9 for the TiffWriter')
            elif compression > 0:
                self.compression = compression
        self.tracker = None
        self.trackerfile = None
        self.trackerFlag = Event()
        self.trackerpar = None

    def close_file(self):
        if not self.fd is None:
            self.fd.close()
        self.fd = None

    def _open_file(self,filename,frame = None):
        self.fd = twriter(filename)

    def _write(self,frame,frameid,timestamp):
        self.fd.save(frame,
                     compress=self.compression,
                     description='id:{0};timestamp:{1}'.format(frameid,
                                                               timestamp))

################################################################################
################################################################################
################################################################################
class BinaryWriter(GenericWriterProcess):
    def __init__(self,
                 cam,
                 loggerQ = None,
                 filename = pjoin('dummy','run'),
                 dataname = 'eyecam',
                 datafolder=pjoin(os.path.expanduser('~'),'data'),
                 pathformat = pjoin('{datafolder}','{dataname}','{filename}',
                                    '{today}_{run}_{nfiles}'),
                 framesperfile=0,
                 sleeptime = 1./300,
                 virtual_channels = None,
                 incrementruns=True,**kwargs):
        self.extension = '_{nchannels}_{W}_{H}_{dtype}.dat'
        super(BinaryWriter,self).__init__(cam = cam,
                                          loggerQ=loggerQ,
                                          filename=filename,
                                          datafolder=datafolder,
                                          dataname=dataname,
                                          pathformat = pathformat,
                                          framesperfile=framesperfile,
                                          sleeptime=sleeptime,
                                          virtual_channels = virtual_channels,
                                          incrementruns=incrementruns,
                                          **kwargs)
        self.buf = []
    def close_file(self):
        if not self.fd is None:
            self.fd.close()
        self.fd = None

    def _open_file(self,filename,frame = None):
        dtype = frame.dtype
        if dtype == np.float32:
            dtype='float32'
        elif dtype == np.uint8:
            dtype='uint8'
        else:
            dtype='uint16'
        filename = filename.format(nchannels = self.virtual_channels.value,
                                   W=self.w.value,
                                   H=self.h.value,
                                   dtype=dtype) 
        self.parsed_filename = filename
        self.fd = open(filename,'wb')
        
    def _write(self,frame,frameid,timestamp):
        self.fd.write(frame)
        if np.mod(frameid,5000) == 0: 
            display('Wrote frame id - {0}'.format(frameid))
        
################################################################################
################################################################################
################################################################################
nvenc_presets = {0:'default (medium)',
                 1:'slow',
                 2:'medium',
                 3:'fast',
                 4:'hp',
                 5:'hq',
                 6:'bd',
                 8:'llhq',
                 9:'llhp',
                 10:'lossless',
                 11:'losslesshp'}
class FFMPEGWriter(GenericWriterProcess):
    def __init__(self,
                 cam,
                 loggerQ = None,
                 filename = pjoin('dummy','run'),
                 dataname = 'eyecam',
                 datafolder=pjoin(os.path.expanduser('~'),'data'),
                 pathformat = pjoin('{datafolder}','{dataname}','{filename}',
                                    '{today}_{run}_{nfiles}'),
                 framesperfile=0,
                 sleeptime = 1./30,
                 incrementruns=True,
                 hwaccel = None,
                 preset = None,
                 compression=None,
                 bitrate = '5M',
                 **kwargs):
        self.extension = '.avi'
        super(FFMPEGWriter,self).__init__(cam = cam,
                                          loggerQ=loggerQ,
                                          filename=filename,
                                          datafolder=datafolder,
                                          dataname=dataname,
                                          pathformat = pathformat,
                                          framesperfile=framesperfile,
                                          sleeptime=sleeptime,
                                          incrementruns=incrementruns,
                                          **kwargs)
        self.compression = compression
        if self.compression is None:
            self.compression = 0
        if type(preset) == str:
            preset =[nvenc_presets[k] for k in  nvenc_presets.keys()].index(preset)
        self.preset = preset
        if hwaccel is None:
            if self.compression == 0:
                self.compression = 25
            self.doutputs = {'-format':'h264',
                             '-pix_fmt':'gray',
                             '-vcodec':'libx264',
                             '-threads':str(10),
                             '-crf':str(self.compression)}
        else:            
            if hwaccel == 'intel':
                if self.compression == 0:
                    self.compression = 25
                self.doutputs = {'-format':'h264',
                                 '-pix_fmt':'yuv420p',#'gray',
                                 '-vcodec':'h264_qsv',#'libx264',
                                 '-global_quality':str(25), # specific to the qsv
                                 '-look_ahead':str(1),
                                 #preset='veryfast',#'ultrafast',
                                 '-threads':str(1),
                                 '-crf':str(self.compression)}
            elif hwaccel == 'nvidia':
                self.doutputs = {'-vcodec':'h264_nvenc',
                                 '-pix_fmt':'yuv420p',
                                 '-tune': 'hq',
                                 '-qmin': '0',
                                 '-g': '250',
                                 '-bf': '3',
                                 '-b_ref_mode':'middle',
                                 '-temporal-aq': '1',
                                 '-rc-lookahead':'20',
                                 '-i_qfactor': '0.75',
                                 '-b_qfactor': '1.1',
                                 '-maxrate':'25M',
                                 '-b:v':bitrate,
                                 '-threads':str(1)}
                preset = 'NA'
                if not self.preset is None:
                    print('Using preset for compression')
                    self.doutputs['-preset'] = str(self.preset)
                    preset = nvenc_presets[self.preset]

                if not self.compression is None:
                    comp = self.compression
                    if type(self.compression) is str:
                        if ':' in self.compression: # then parse the bitrate
                            comp = self.compression.split(':')
                        else:
                            comp = [str(self.compression)]
                        self.compression = comp[0]
                        self.doutputs['-cq:v'] = comp[0]
                        #self.doutputs['-rc:v'] = 'vbr_hq'
                        if len(comp)>1:
                            self.doutputs['-b:v'] = comp[1]
                            bitrate = comp[1]
        self.hwaccel = hwaccel
        display('[FFMPEG] - Using compression (preset {0}) {1}  - bitrate {3} for the {2} encoder.'.format(
            preset,
            self.compression,
            hwaccel,
            bitrate))
        
    def close_file(self):
        if not self.fd is None:
            self.fd.close()
        self.fd = None

    def _open_file(self,filename,frame = None):
        if frame is None:
            raise ValueError('[Recorder] Need to pass frame to open a file.')
        self.frame_rate = self.cam['frame_rate'].value
        if self.frame_rate is None or self.frame_rate == 0:
            display('Using 30Hz frame rate for ffmpeg')
            self.frame_rate = 30
        
        self.doutputs['-r'] =str(self.frame_rate)
        self.dinputs = {'-r':str(self.frame_rate)}

        # does a check for the datatype, if uint16 then save compressed lossless
        if frame.dtype in [np.uint16] and len(frame.shape) == 2:
            self.fd = FFmpegWriter(filename.replace(self.extension,'.mov'),
                                   inputdict={'-pix_fmt':'gray16le',
                                              '-r':str(self.frame_rate)}, # this is important
                                   outputdict={'-c:v':'libopenjpeg',
                                               '-pix_fmt':'gray16le',
                                               '-r':str(self.frame_rate)})
            return
        elif len(frame.shape) == 3 and (frame.shape[-1] == 3):
            self.doutputs['-pix_fmt'] = 'yuv420p'
            display('Camera has 3 channels; recording in yuv420p.')
        self.fd = FFmpegWriter(filename,
                               inputdict=self.dinputs,
                               outputdict=self.doutputs)
            
    def _write(self,frame,frameid,timestamp):
        self.fd.writeFrame(frame)

################################################################################
################################################################################
################################################################################
class OpenCVWriter(GenericWriter):
    def __init__(self,
                 cam,
                 loggerQ = None,
                 filename = pjoin('dummy','run'),
                 dataname = 'eyecam',
                 pathformat = pjoin('{datafolder}','{dataname}','{filename}',
                                    '{today}_{run}_{nfiles}'),
                 datafolder=pjoin(os.path.expanduser('~'),'data'),
                 framesperfile=0,
                 sleeptime = 1./30,
                 incrementruns=True,
                 compression=None,
                 fourcc = 'X264',**kwargs):
        self.extension = '.avi'
        super(OpenCVWriter,self).__init__(cam = cam,
                                          loggerQ=loggerQ,
                                          filename=filename,
                                          datafolder=datafolder,
                                          pathformat = pathformat,
                                          dataname=dataname,
                                          framesperfile=framesperfile,
                                          sleeptime=sleeptime,
                                          incrementruns=incrementruns,
                                          **kwargs)
        cv2.setNumThreads(6)
        self.compression = 17
        if not compression is None:
            if compression > 0:
                self.compression = compression
        self.fourcc = cv2.VideoWriter_fourcc(*fourcc)
        
    def close_file(self):
        if not self.fd is None:
            self.fd.release()
        self.fd = None

    def _open_file(self,filename,frame = None):
        self.isColor = False
        if len(frame.shape) < 2:
            self.isColor = True
        self.fd = cv2.VideoWriter(filename,
                                  cv2.CAP_FFMPEG,#cv2.CAP_DSHOW,#cv2.CAP_INTEL_MFX,
                                  self.fourcc,120,(self.w.value,self.h.value),self.isColor)

    def _write(self,frame,frameid,timestamp):
        if len(frame.shape) < 2:
            frame = cv2.cvtColor(frame,cv2.COLOR_GRAY2RGB)
        self.fd.write(frame)

        
################################################################################
################################################################################
################################################################################

class FFMPEGCamWriter(GenericWriter):
    def __init__(self,
                 cam,
                 filename = pjoin('dummy','run'),
                 dataname = 'eyecam',
                 datafolder=pjoin(os.path.expanduser('~'),'data'),
                 pathformat = pjoin('{datafolder}','{dataname}','{filename}',
                                    '{today}_{run}_{nfiles}'),
                 framesperfile=0,
                 incrementruns=True,
                 compression=None,
                 hwaccel = None,**kwargs):
        self.extension = '.avi'
        self.nchannels = cam.nchan

        super(FFMPEGCamWriter,self).__init__(cam = cam,
                                             filename=filename,
                                             datafolder=datafolder,
                                             dataname=dataname,
                                             pathformat = pathformat,
                                             framesperfile=framesperfile,
                                             incrementruns=incrementruns,
                                             **kwargs)

        self.compression = compression
        if self.compression is None:
            self.compression = 0
        if hwaccel is None:
            self.doutputs = {'-format':'h264',
                             '-pix_fmt':'gray',
                             '-vcodec':'libx264',
                             '-threads':str(10),
                             '-crf':str(self.compression)}
        else:            
            if hwaccel == 'intel':
                if self.compression == 0:
                    self.compression = 17
                self.doutputs = {'-format':'h264',
                                 '-pix_fmt':'yuv420p',#'gray',
                                 '-vcodec':'h264_qsv',#'libx264',
                                 '-global_quality':str(25), # specific to the qsv
                                 '-look_ahead':str(1),
                                 #preset='veryfast',#'ultrafast',
                                 '-threads':str(1),
                                 '-crf':str(self.compression)}
            elif hwaccel == 'nvidia':
                if self.compression == 0:
                    self.compression = 25
                self.doutputs = {'-vcodec':'h264_nvenc',
                                 '-pix_fmt':'yuv420p',
                                 #'-cq:v':str(self.compression),
                                 '-threads':str(1),
                                 '-preset':str(self.compression)}
        self.hwaccel = hwaccel
        display('Using compression {0} for the {1} FFMPEG encoder.'.format(
            self.compression, hwaccel))

    def close_file(self):
        if not self.fd is None:
            self.fd.close()
            print("------->>> Closed file.")
        self.fd = None

    def _open_file(self,filename,frame = None):
        if frame is None:
            raise ValueError('[Recorder] Need to pass frame to open a file.')
        self.frame_rate = self.cam['frame_rate'].value
        if self.frame_rate is None or self.frame_rate == 0:
            display('Using 30Hz frame rate for ffmpeg')
            self.frame_rate = 30
        
        self.doutputs['-r'] =str(self.frame_rate)
        self.dinputs = {'-r':str(self.frame_rate)}

        # does a check for the datatype, if uint16 then save compressed lossless
        if frame.dtype in [np.uint16] and len(frame.shape) == 2:
            self.fd = FFmpegWriter(filename.replace(self.extension,'.mov'),
                                   inputdict={'-pix_fmt':'gray16le',
                                              '-r':str(self.frame_rate)}, # this is important
                                   outputdict={'-c:v':'libopenjpeg',
                                               '-pix_fmt':'gray16le',
                                               '-r':str(self.frame_rate)})
        else:
            self.fd = FFmpegWriter(filename,
                                   inputdict=self.dinputs,
                                   outputdict=self.doutputs)
            
    def _write(self,frame,frameid,timestamp):
        self.fd.writeFrame(frame)

################################################################################
################################################################################
################################################################################

class BinaryCamWriter(GenericWriter):
    def __init__(self,
                 cam,
                 filename = pjoin('dummy','run'),
                 dataname = 'eyecam',
                 datafolder=pjoin(os.path.expanduser('~'),'data'),
                 pathformat = pjoin('{datafolder}','{dataname}','{filename}',
                                    '{today}_{run}_{nfiles}'),
                 framesperfile=0,
                 virtual_channels = None,
                 incrementruns=True,**kwargs):
        self.extension = '_{nchannels}_{H}_{W}_{dtype}.dat'
        super(BinaryCamWriter,self).__init__(cam = cam,
                                             filename=filename,
                                             datafolder=datafolder,
                                             dataname=dataname,
                                             pathformat = pathformat,
                                             virtual_channels = virtual_channels,
                                             framesperfile=framesperfile,
                                             incrementruns=incrementruns,
                                             **kwargs)                

    def close_file(self):
        if not self.fd is None:
            self.fd.close()
            print("------->>> Closed file.")
        self.fd = None

    def _open_file(self,filename,frame = None):
        dtype = frame.dtype
        if dtype == np.float32:
            dtype='float32'
        elif dtype == np.uint8:
            dtype='uint8'
        else:
            dtype='uint16'
        filename = filename.format(nchannels = self.virtual_channels.value,
                                   W=self.w.value,
                                   H=self.h.value, dtype=dtype)
        self.parsed_filename = filename
        self.fd = open(filename,'wb')

    def _write(self,frame,frameid,timestamp):
        self.fd.write(frame)

class BinaryDAQWriter(GenericWriter):
    def __init__(self,
                 daq,
                 filename = pjoin('dummy','run'),
                 dataname = 'daq',
                 datafolder=pjoin(os.path.expanduser('~'),'data'),
                 pathformat = pjoin('{datafolder}','{dataname}','{filename}',
                                    '{today}_{run}_{nfiles}'),
                 incrementruns=True,**kwargs):
        self.extension = '_{nchannels}_{dtype}.nidq.bin'
        self.daq = daq
        super(BinaryDAQWriter,self).__init__(cam = daq,
                                             filename=filename,
                                             datafolder=datafolder,
                                             dataname=dataname,
                                             pathformat = pathformat,
                                             framesperfile=-1,
                                             incrementruns=incrementruns,
                                             **kwargs)
        self.nchannels = daq.ai_num_channels + daq.di_num_channels
        self.nbytes = 0
    def close_file(self):
        if not self.fd is None:
            self.fd.close()
            print("------->>> Closed nidq.bin file.")
            # create the metadata:
            with open(self.parsed_filename.replace('.bin','.meta'),'w') as f:
                f.write('fileSizeBytes={0}\n'.format(self.nbytes))
                f.write('fileTimeSecs={0}\n'.format(self.nsamples/self.daq.srate))
                f.write('niSampRate={0}\n'.format(self.daq.srate))
                f.write('niAiRangeMax={0}\n'.format(self.daq.ai_range[1]))
                f.write('niAiRangeMin={0}\n'.format(self.daq.ai_range[0]))
                f.write('niMAGain={0}\n'.format(1))
                f.write('niMNGain={0}\n'.format(1))
                f.write('snsMnMaXaDw=0,0,{0},{1}\n'.format(self.daq.ai_num_channels,
                                                            self.daq.ai_num_channels))                
                f.write('nSavedChans={0}\n'.format(self.nchannels))
                f.write('typeThis=nidq\n')
                p = []
                for k in self.daq.digital_channels.keys():
                    p.append('({0},{1})'.format(k,self.daq.digital_channels[k]))
                f.write('digitalChannelNames={0}\n'.format(','.join(p)))
                p = []
                for k in self.daq.analog_channels.keys():
                    p.append('({0},{1})'.format(k,self.daq.analog_channels[k]))
                f.write('analogChannelNames={0}\n'.format(','.join(p)))
                
        self.fd = None

    def open_file(self,nfiles = None,data = None):
        filename = self.get_filename_path()
        if not self.fd is None:
            self.close_file()
        self._open_file(filename,data)
        self.nFiles += 1
        if hasattr(self,'parsed_filename'):
            filename = self.parsed_filename
        display('Opened: '+ filename)

    def _open_file(self,filename,data = None):
        dtype = data.dtype
        if dtype == np.float32:
            dtype='float32'
        elif dtype == np.int16:
            dtype='int16'
        elif dtype == np.uint8:
            dtype='uint8'
        else:
            dtype='uint16'
        filename = filename.format(nchannels = self.nchannels,
                                   dtype=dtype)
        self.parsed_filename = filename
        self.fd = open(filename,'wb')
        self.nsamples = 0
        self.nbytes = 0

    def save(self, data,metadata = None):
        if self.fd is None:
            self.open_file(data = data)
        return self._write(data)
    
    def _write(self,data):
        self.fd.write(data)
        self.nbytes += data.nbytes
        self.nsamples += data.shape[1]
        
    def close_run(self):
        self.close_file()
        self.runs += 1
    
class TiffCamWriter(GenericWriter):
    def __init__(self,
                 cam,
                 filename = pjoin('dummy','run'),
                 dataname = 'cam',
                 datafolder=pjoin(os.path.expanduser('~'),'data'),
                 pathformat = pjoin('{datafolder}','{dataname}','{filename}',
                                    '{today}_{run}_{nfiles}'),
                 framesperfile=256,
                 sleeptime = 1./300,
                 incrementruns=True,
                 compression = None,**kwargs):
        self.extension = '.tif'
        super(TiffCamWriter,self).__init__(cam = cam,
                                           datafolder=datafolder,
                                           filename=filename,
                                           dataname=dataname,
                                           pathformat=pathformat,
                                           framesperfile=framesperfile,
                                           sleeptime=sleeptime,
                                           incrementruns=incrementruns,
                                           **kwargs)
        self.compression = None
        if not compression is None:
            self.compression = np.clip(compression,0,9)
            
    def close_file(self):
        if not self.fd is None:
            self.fd.close()
            print("------->>> Closed file.")
        self.fd = None

    def _open_file(self,filename,frame = None):
        self.fd = twriter(filename)

    def _write(self,frame,frameid,timestamp):
        self.fd.save(frame,
                     compress=self.compression,
                     description='id:{0};timestamp:{1}'.format(frameid,timestamp))
        
################################################################################
################################################################################
################################################################################

def parseCamLog(fname, readTeensy = False):
    logheaderkey = '# Log header:'
    comments = []
    with open(fname,'r',encoding = 'utf-8') as fd:
        for line in fd:
            if line.startswith('#'):
                line = line.strip('\n').strip('\r')
                comments.append(line)
                if line.startswith(logheaderkey):
                    columns = line.strip(logheaderkey).strip(' ').split(',')

    logdata = pd.read_csv(fname, 
                          delimiter=',',
                          header=None,
                          comment='#',
                          engine='c')
    col = [c for c in logdata.columns]
    for icol in range(len(col)):
        if icol <= len(columns)-1:
            col[icol] = columns[icol]
        else:
            col[icol] = 'var{0}'.format(icol)
    logdata.columns = col
    if readTeensy:
        # get the sync pulses and frames along with the LED
        def _convert(string):
            try:
                val = int(string)
            except ValueError as err:
                val = float(string)
            return val

        led = []
        sync= []
        ncomm = []
        for l in comments:
            if l.startswith('#LED:'):
                led.append([_convert(f) for f in  l.strip('#LED:').split(',')])
            elif l.startswith('#SYNC:'):
                sync.append([0.] + [_convert(f) for f in  l.strip('#SYNC:').split(',')])
            elif l.startswith('#SYNC1:'):
                sync.append([1.] + [_convert(f) for f in  l.strip('#SYNC1:').split(',')])
            else:
                ncomm.append(l)
        if len(sync):
            sync = pd.DataFrame(sync,
                                columns=['sync','count','frame','timestamp'])
        if len(led):
            led = pd.DataFrame(led, columns=['led','frame','timestamp'])
        return logdata,led,sync,ncomm
    return logdata,comments

parse_cam_log = parseCamLog

class TiffStack(object):
    def __init__(self,filenames):
        if type(filenames) is str:
            filenames = np.sort(glob(pjoin(filenames,'*.tif')))
        
        assert type(filenames) in [list,np.ndarray], 'Pass a list of filenames.'
        self.filenames = filenames
        for f in filenames:
            assert os.path.exists(f), f + ' not found.'
        # Get an estimate by opening only the first and last files
        framesPerFile = []
        self.files = []
        for i,fn in enumerate(self.filenames):
            if i == 0 or i == len(self.filenames)-1:
                self.files.append(TiffFile(fn))
            else:
                self.files.append(None)                
            f = self.files[-1]
            if i == 0:
                dims = f.series[0].shape
                self.shape = dims
            elif i == len(self.filenames)-1:
                dims = f.series[0].shape
            framesPerFile.append(np.int64(dims[0]))
        self.framesPerFile = np.array(framesPerFile, dtype=np.int64)
        self.framesOffset = np.hstack([0,np.cumsum(self.framesPerFile[:-1])])
        self.nFrames = np.sum(framesPerFile)
        self.curfile = 0
        self.curstack = self.files[self.curfile].asarray()
        N,self.h,self.w = self.curstack.shape[:3]
        self.dtype = self.curstack.dtype
        self.shape = (self.nFrames,self.shape[1],self.shape[2])
    def getFrameIndex(self,frame):
        '''Computes the frame index from multipage tiff files.'''
        fileidx = np.where(self.framesOffset <= frame)[0][-1]
        return fileidx,frame - self.framesOffset[fileidx]
    def __getitem__(self,*args):
        index  = args[0]
        if not type(index) is int:
            Z, X, Y = index
            if type(Z) is slice:
                index = range(Z.start, Z.stop, Z.step)
            else:
                index = Z
        else:
            index = [index]
        img = np.empty((len(index),self.h,self.w),dtype = self.dtype)
        for i,ind in enumerate(index):
            img[i,:,:] = self.getFrame(ind)
        return np.squeeze(img)
    def getFrame(self,frame):
        ''' Returns a single frame from the stack '''
        fileidx,frameidx = self.getFrameIndex(frame)
        if not fileidx == self.curfile:
            if self.files[fileidx] is None:
                self.files[fileidx] = TiffFile(self.filenames[fileidx])
            self.curstack = self.files[fileidx].asarray()
            self.curfile = fileidx
        return self.curstack[frameidx,:,:]
    def __len__(self):
        return self.nFrames

def mmap_dat(filename,
             mode = 'r',
             nframes = None,
             shape = None,
             dtype='uint16'):
    '''
    Loads frames from a binary file as a memory map.
    This is useful when the data does not fit to memory.
    
    Inputs:
        filename (str)       : fileformat convention, file ends in _NCHANNELS_H_W_DTYPE.dat
        mode (str)           : memory map access mode (default 'r')
                'r'   | Open existing file for reading only.
                'r+'  | Open existing file for reading and writing.                 
        nframes (int)        : number of frames to read (default is None: the entire file)
        shape (list|tuple)   : dimensions (NCHANNELS, HEIGHT, WIDTH) default is None
        dtype (str)          : datatype (default uint16) 
    Returns:
        A memory mapped  array with size (NFRAMES,[NCHANNELS,] HEIGHT, WIDTH).

    Example:
        dat = mmap_dat(filename)

    Joao Couto - from wfield
    '''
    
    if not os.path.isfile(filename):
        raise OSError('File {0} not found.'.format(filename))
    if shape is None or dtype is None: # try to get it from the filename
        meta = os.path.splitext(filename)[0].split('_')
        if shape is None:
            try: # Check if there are multiple channels
                shape = [int(m) for m in meta[-4:-1]]
            except ValueError:
                shape = [int(m) for m in meta[-3:-1]]
        if dtype is None:
            dtype = meta[-1]
    dt = np.dtype(dtype)
    if nframes is None:
        # Get the number of samples from the file size
        nframes = int(os.path.getsize(filename)/(np.prod(shape)*dt.itemsize))
    dt = np.dtype(dtype)
    return np.memmap(filename,
                     mode=mode,
                     dtype=dt,
                     shape = (int(nframes),*shape))


def stack_to_mj2_lossless(stack,fname, rate = 30):
    '''
    Compresses a uint16 stack with FFMPEG and libopenjpeg
    
    Inputs:
        stack                : array or memorymapped binary file
        fname                : output filename (will change extension to .mov)
        rate                 : rate of the mj2 movie [30 Hz default]

    Example:
       from labcams.io import * 
       fname = '20200710_140729_2_540_640_uint16.dat'
       stack = mmap_dat(fname)
       stack_to_mj2_lossless(stack,fname, rate = 30)
    '''
    ext = os.path.splitext(fname)[1]
    assert len(ext), "[mj2 conversion] Need to pass a filename {0}.".format(fname)
    
    if not ext == '.mov':
        print('[mj2 conversion] Changing extension to .mov')
        outfname = fname.replace(ext,'.mov')
    else:
        outfname = fname
    assert stack.dtype == np.uint16, "[mj2 conversion] This only works for uint16 for now."

    nstack = stack.reshape([-1,*stack.shape[2:]]) # flatten if needed    
    sq = FFmpegWriter(outfname, inputdict={'-pix_fmt':'gray16le',
                                              '-r':str(rate)}, # this is important
                      outputdict={'-c:v':'libopenjpeg',
                                  '-pix_fmt':'gray16le',
                                  '-r':str(rate)})
    from tqdm import tqdm
    for i,f in tqdm(enumerate(nstack),total=len(nstack)):
        sq.writeFrame(f)
    sq.close()
    
